"""This module contains the Reporter class, which is used to report data to disk."""

from __future__ import annotations

import datetime
import re
from operator import attrgetter
from pathlib import Path
from typing import TYPE_CHECKING, Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import zarr.storage
from dateutil.relativedelta import relativedelta
from xarray.backends.zarr import FillValueCoder
from zarr.codecs import ZstdCodec

from geb.geb_types import ArrayFloat32, ArrayFloat64, ArrayInt64, TwoDArrayInt32
from geb.module import Module
from geb.store import DynamicArray
from geb.workflows.io import fast_rmtree, read_geom
from geb.workflows.methods import multi_level_merge
from geb.workflows.raster import coord_to_pixel

if TYPE_CHECKING:
    from geb.model import GEBModel

WATER_CIRCLE_REPORT_CONFIG: dict[str, str | dict[str, str | dict[str, str]]] = {
    "hydrology": {
        "_water_circle_storage": {
            "varname": ".current_storage",
            "type": "scalar",
        },
        "_water_circle_routing_loss": {
            "varname": ".routing_loss_m3",
            "type": "scalar",
        },
    },
    "hydrology.landsurface": {
        "_water_circle_rain": {
            "varname": ".rain_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_circle_snow": {
            "varname": ".snow_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_circle_transpiration": {
            "varname": ".transpiration_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_circle_bare_soil_evaporation": {
            "varname": ".bare_soil_evaporation_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_circle_open_water_evaporation": {
            "varname": ".open_water_evaporation_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_circle_interception_evaporation": {
            "varname": ".interception_evaporation_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_circle_sublimation_or_deposition": {
            "varname": ".sublimation_or_deposition_m",
            "type": "HRU",
            "function": "weightedsum",
        },
    },
    "hydrology.routing": {
        "_water_circle_river_evaporation": {
            "varname": ".total_evaporation_in_rivers_m3",
            "type": "scalar",
        },
        "_water_circle_waterbody_evaporation": {
            "varname": ".total_waterbody_evaporation_m3",
            "type": "scalar",
        },
        "_water_circle_river_outflow": {
            "varname": ".total_outflow_at_pits_m3",
            "type": "scalar",
        },
    },
    "hydrology.water_demand": {
        "_water_circle_domestic_water_loss": {
            "varname": ".domestic_water_loss_m3",
            "type": "scalar",
        },
        "_water_circle_industry_water_loss": {
            "varname": ".industry_water_loss_m3",
            "type": "scalar",
        },
        "_water_circle_livestock_water_loss": {
            "varname": ".livestock_water_loss_m3",
            "type": "scalar",
        },
    },
}

WATER_BALANCE_REPORT_CONFIG = {
    "hydrology": {
        "_water_balance_storage": {
            "varname": ".current_storage",
            "type": "scalar",
        },
        "_water_balance_routing_loss": {
            "varname": ".routing_loss_m3",
            "type": "scalar",
        },
    },
    "hydrology.landsurface": {
        "_water_balance_rain": {
            "varname": ".rain_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_balance_snow": {
            "varname": ".snow_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_balance_transpiration": {
            "varname": ".transpiration_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_balance_bare_soil_evaporation": {
            "varname": ".bare_soil_evaporation_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_balance_open_water_evaporation": {
            "varname": ".open_water_evaporation_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_balance_interception_evaporation": {
            "varname": ".interception_evaporation_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_balance_sublimation_or_deposition": {
            "varname": ".sublimation_or_deposition_m",
            "type": "HRU",
            "function": "weightedsum",
        },
        "_water_balance_interflow": {
            "varname": ".interflow_m",
            "type": "HRU",
            "function": "weightedsum",
        },
    },
    "hydrology.routing": {
        "_water_balance_river_evaporation": {
            "varname": ".total_evaporation_in_rivers_m3",
            "type": "scalar",
        },
        "_water_balance_waterbody_evaporation": {
            "varname": ".total_waterbody_evaporation_m3",
            "type": "scalar",
        },
        "_water_balance_river_outflow": {
            "varname": ".total_outflow_at_pits_m3",
            "type": "scalar",
        },
    },
    "hydrology.water_demand": {
        "_water_balance_domestic_water_loss": {
            "varname": ".domestic_water_loss_m3",
            "type": "scalar",
        },
        "_water_balance_industry_water_loss": {
            "varname": ".industry_water_loss_m3",
            "type": "scalar",
        },
        "_water_balance_livestock_water_loss": {
            "varname": ".livestock_water_loss_m3",
            "type": "scalar",
        },
    },
}

ENERGY_BALANCE_REPORT_CONFIG = {
    "hydrology.landsurface": {
        "_energy_balance_soil_temperature_layer_0_C": {
            "varname": ".soil_temperature_C[0]",
            "type": "HRU",
            "function": "weightedmean",
        },
        "_energy_balance_soil_temperature_layer_1_C": {
            "varname": ".soil_temperature_C[1]",
            "type": "HRU",
            "function": "weightedmean",
        },
        "_energy_balance_soil_temperature_layer_2_C": {
            "varname": ".soil_temperature_C[2]",
            "type": "HRU",
            "function": "weightedmean",
        },
        "_energy_balance_soil_temperature_layer_3_C": {
            "varname": ".soil_temperature_C[3]",
            "type": "HRU",
            "function": "weightedmean",
        },
        "_energy_balance_soil_temperature_layer_4_C": {
            "varname": ".soil_temperature_C[4]",
            "type": "HRU",
            "function": "weightedmean",
        },
        "_energy_balance_soil_temperature_layer_5_C": {
            "varname": ".soil_temperature_C[5]",
            "type": "HRU",
            "function": "weightedmean",
        },
    },
}


def get_fill_value(
    data: np.ndarray[Any] | DynamicArray,
) -> tuple[int | float | None, Any]:
    """Get the fill value for a zarr array based on the data type.

    Args:
        data: The data array.

    Returns:
        A tuple containing the fill value and the encoded fill value for the zarr array.

    Raises:
        ValueError: If the data type is not recognized.
    """
    if np.issubdtype(data.dtype, np.floating):
        kind = np.dtype(data.dtype)
        fill_value: int | float = np.nan
    elif np.issubdtype(data.dtype, np.integer):
        kind = np.dtype(data.dtype)
        fill_value = -1
    elif data.dtype == bool:
        fill_value = None
    else:
        raise ValueError(f"Value dtype {data.dtype} not recognized.")

    if fill_value is not None:
        encoded_fill_value = FillValueCoder.encode(fill_value, kind)
    else:
        encoded_fill_value = None

    return fill_value, encoded_fill_value


def get_time_chunk_size(
    dtype: np.dtype, *other_dim_sizes: int, target_size_bytes: int = 100_000_000
) -> int:
    """Get the time chunk size for a zarr array.

    Args:
        dtype: The data type of the array.
        other_dim_sizes: The sizes of the other dimensions of the array.
        target_size_bytes: The target size of the chunk in bytes.

    Returns:
        time_chunk_size: The size of the time chunk.
    """
    type_size: int = np.dtype(dtype).itemsize
    other_dims_size: np.int64 = np.prod(other_dim_sizes)
    time_chunk_size: int = (target_size_bytes // (type_size * other_dims_size)).item()
    return max(1, time_chunk_size)


def create_time_array(
    start: datetime.datetime,
    end: datetime.datetime,
    timestep: datetime.timedelta | relativedelta,
    conf: dict,
    substeps: None | int = None,
) -> ArrayInt64:
    """Create a time array based on the start and end time, the timestep, and the frequency.

    Args:
        start: The start time.
        end: The end time.
        timestep: The timestep length.
        conf: The configuration for the frequency.
        substeps: The number of substeps per timestep.

    Returns:
        time: The time array.

    Raises:
        ValueError: If the frequency is not recognized.
        ValueError: If substeps are provided for a frequency that does not support them.
    """
    if "frequency" not in conf:
        frequency = {"every": "day"}
    else:
        frequency = conf["frequency"]
    if "every" in frequency:
        every = frequency["every"]
        time = []
        current_time = start
        while current_time <= end:
            if every == "year":
                if substeps is not None:
                    raise ValueError(
                        "Substeps not supported for yearly frequency in create_time_array."
                    )
                if (
                    frequency["month"] == current_time.month
                    and frequency["day"] == current_time.day
                ):
                    time.append(current_time)
            elif every == "month":
                if substeps is not None:
                    raise ValueError(
                        "Substeps not supported for monthly frequency in create_time_array."
                    )
                if frequency["day"] == current_time.day:
                    time.append(current_time)
            elif every == "day":
                if substeps is None:
                    time.append(current_time)
                else:
                    for substep in range(substeps):
                        time.append(current_time + substep * (timestep / substeps))
            current_time += timestep
    elif frequency == "initial":
        if substeps is not None:
            raise ValueError(
                "Substeps not supported for initial frequency in create_time_array."
            )
        time = [start]
    elif frequency == "final":
        if substeps is not None:
            raise ValueError(
                "Substeps not supported for final frequency in create_time_array."
            )
        time = [end]
    else:
        raise ValueError(f"Frequency {frequency} not recognized.")

    time_array = (
        np.array(time, dtype="datetime64[ns]").astype("datetime64[s]").astype(np.int64)
    )
    return time_array


def prepare_gridded_group(
    name: str,
    config: dict,
    lon: ArrayFloat32 | ArrayFloat64,
    lat: ArrayFloat32 | ArrayFloat64,
    crs: str,
    example_value: np.ndarray[Any],
    time: ArrayInt64,
    chunk_size: int,
    compression_level: int,
) -> None:
    """Create a zarr group for gridded data.

    Args:
        name: The name of the variable.
        config: The configuration for the variable.
        lon: The longitude array.
        lat: The latitude array.
        crs: The coordinate reference system in WKT format.
        example_value: An example value to determine the data type.
        time: The time array.
        chunk_size: The size of the time chunk.
        compression_level: The compression level for the zarr array.
    """
    root_group = config["_root_group"]

    # Create the y coordinate array
    y_group = root_group.create_array(
        "y",
        shape=lat.shape,
        dtype=lat.dtype,
        dimension_names=["y"],
        attributes={
            "standard_name": "latitude",
            "units": "degrees_north",
        },
    )
    y_group[:] = lat

    # Create the x coordinate array
    x_group = root_group.create_array(
        "x",
        shape=lon.shape,
        dtype=lon.dtype,
        dimension_names=["x"],
        attributes={
            "standard_name": "longitude",
            "units": "degrees_east",
        },
    )
    x_group[:] = lon

    # Create the time coordinate array
    time_group = root_group.create_array(
        "time",
        shape=time.shape,
        dtype=time.dtype,
        dimension_names=["time"],
        attributes={
            "standard_name": "time",
            "units": "seconds since 1970-01-01T00:00:00",
            "calendar": "gregorian",
        },
    )
    time_group[:] = time

    fill_value, encoded_fill_value = get_fill_value(example_value)

    attributes = {
        "grid_mapping": "crs",
        "coordinates": "time y x",
        "units": "unknown",
        "long_name": name,
        "_CRS": {"wkt": crs},
    }
    if encoded_fill_value is not None:
        attributes["_FillValue"] = encoded_fill_value

    # Create the variable array
    variable_data: zarr.Array[Any] = root_group.create_array(
        name,
        shape=(
            time_group.size,
            lat.size,
            lon.size,
        ),
        chunks=(
            chunk_size,
            lat.size,
            lon.size,
        ),
        dtype=example_value.dtype,
        compressors=(
            ZstdCodec(
                level=compression_level,
            ),
        ),
        fill_value=fill_value,
        dimension_names=["time", "y", "x"],
        attributes=attributes,
    )

    # Pre-allocate the writing buffer for the chunks
    config["_chunk_data"] = np.full(
        (chunk_size, lat.size, lon.size),
        np.nan,
        dtype=np.float32,
    )


def prepare_agent_group(
    name: str,
    config: dict,
    time: ArrayInt64,
    example_value: np.ndarray[Any] | int | float | bool,
    chunk_target_size_bytes: int,
    compression_level: int,
) -> None:
    """Create a zarr group for agent data.

    Args:
        name: The name of the variable.
        config: The configuration for the variable.
        time: The time array.
        example_value: An example value to determine the data type.
        chunk_target_size_bytes: The target size of the chunk in bytes.
        compression_level: The compression level for the zarr array.
    """
    root_group = config["_root_group"]

    # Create the time coordinate array
    time_group = root_group.create_array(
        "time",
        shape=time.shape,
        dtype=time.dtype,
        dimension_names=["time"],
        attributes={
            "standard_name": "time",
            "units": "seconds since 1970-01-01T00:00:00",
            "calendar": "gregorian",
        },
    )
    time_group[:] = time

    # Determine chunk size and shape based on example value
    if isinstance(example_value, (float, int)):
        shape = (time_group.size,)
        chunk_size = get_time_chunk_size(
            np.dtype(type(example_value)),
            1,
            target_size_bytes=chunk_target_size_bytes,
        )
        if isinstance(example_value, float):
            dtype_ = np.float32
            example_value = np.array(example_value, dtype=np.float32)
        else:
            dtype_ = np.int32
            example_value = np.array(example_value, dtype=np.int32)
        chunk_size = min(chunk_size, time_group.size)
        chunks = (chunk_size,)
        compressors = None
        array_dimensions = ["time"]
    else:
        shape = (time_group.size, example_value.size)
        dtype_ = example_value.dtype
        chunk_size = get_time_chunk_size(
            np.dtype(example_value.dtype),
            example_value.size,
            target_size_bytes=chunk_target_size_bytes,
        )
        chunk_size = min(chunk_size, time_group.size)
        chunks = (chunk_size, example_value.size)
        compressors = (
            ZstdCodec(
                level=compression_level,
            ),
        )
        array_dimensions = ["time", "agents"]

    fill_value, encoded_fill_value = get_fill_value(example_value)
    attributes = {
        "coordinates": "time",
        "units": "unknown",
        "long_name": name,
    }
    if encoded_fill_value is not None:
        attributes["_FillValue"] = encoded_fill_value

    root_group.create_array(
        name,
        shape=shape,
        chunks=chunks,
        dtype=dtype_,
        compressors=compressors,
        fill_value=fill_value,
        dimension_names=array_dimensions,
        attributes=attributes,
    )
    # Pre-allocate the writing buffer for the chunks
    if isinstance(example_value, (float, int)):
        config["_chunk_data"] = np.full((chunk_size,), fill_value)
    else:
        config["_chunk_data"] = np.full((chunk_size, example_value.size), fill_value)


class Reporter:
    """This class is used to report data to disk."""

    def __init__(self, model: GEBModel, clean: bool) -> None:
        """The constructor for the Reporter class.

        Loops over the reporter configuration and creates the necessary files and data structures,
        that are then used while the model is run to add data to.

        For full documentation of the report configuration, see the documentation.

        There are also several pre-defined report configurations that can be activated by adding
        special keys to the report configuration. These are:
        - _discharge_stations: if set to True, discharge at all discharge stations is reported.
        - _outflow_points: if set to True, outflow at all outflow points is reported.
        - _water_circle: if set to True, a standard set of variables to monitor the water circle is reported.
        - _water_balance: if set to True, a standard set of variables to monitor the water balance is reported.
        - _energy_balance: if set to True, a standard set of variables to monitor the energy balance is reported.

        Args:
            model: The GEB model instance.
            clean: If True, the report folder is cleaned at the start of the model run.

        Raises:
            ValueError: If the variable type is not recognized.
        """
        self.model = model
        if "_config" not in self.model.config["report"]:
            self.config: dict[str, int] = {}
        else:
            self.config: dict[str, int] = self.model.config["report"]["_config"].copy()
            del self.model.config["report"]["_config"]

        if self.model.simulate_hydrology:
            self.hydrology = model.hydrology
        self.report_folder = self.model.output_folder / "report" / self.model.run_name
        # optionally clean report model at start of run
        if clean and self.report_folder.exists():
            fast_rmtree(self.report_folder)
        self.report_folder.mkdir(parents=True, exist_ok=True)

        self.variables = {}
        self.timesteps = []

        if (
            self.model.mode == "w"
            and "report" in self.model.config
            and self.model.config["report"]
        ):
            self.is_activated = True

            report_config: dict[str, Any] = self.model.config["report"]

            to_delete: list[str] = []
            for module_name, module_values in list(report_config.items()):
                if module_name.startswith("_"):
                    if module_name == "_discharge_stations" and module_values is True:
                        stations = read_geom(
                            self.model.files["geom"][
                                "discharge/discharge_snapped_locations"
                            ]
                        )

                        station_reporters = {}
                        for station_ID, station_info in stations.iterrows():
                            xy_grid = station_info["snapped_grid_pixel_xy"]
                            station_reporters[
                                f"discharge_hourly_m3_per_s_{station_ID}"
                            ] = {
                                "varname": f"grid.var.discharge_m3_s_per_substep",
                                "type": "grid",
                                "function": f"sample_xy,{xy_grid[0]},{xy_grid[1]}",
                                "substeps": 24,
                            }
                        report_config = multi_level_merge(
                            report_config,
                            {"hydrology.routing": station_reporters},
                        )
                    elif module_name == "_outflow_points" and module_values is True:
                        routing = self.model.hydrology.routing
                        outflow_rivers = routing.outflow_rivers
                        all_rivers = routing.rivers

                        outflow_reporters = {}

                        def get_upstream_represented_xys(
                            river_id: int,
                        ) -> list[tuple[int, int]]:
                            """Recursively find the nearest represented upstream rivers.

                            Args:
                                river_id: The ID of the river to find the upstream represented rivers for.

                            Returns:
                                A list of tuples containing the grid pixel coordinates of the nearest represented upstream rivers.
                            """
                            river = all_rivers.loc[river_id]
                            if river["represented_in_grid"]:
                                return [river["hydrography_xy"][-1]]

                            upstream_rivers = all_rivers[
                                all_rivers["downstream_ID"] == river_id
                            ]
                            xys = []
                            for idx, _ in upstream_rivers.iterrows():
                                xys.extend(get_upstream_represented_xys(idx))
                            return xys

                        for river_ID, river in outflow_rivers.iterrows():
                            assert isinstance(river_ID, int)
                            xys = get_upstream_represented_xys(river_ID)
                            for i, xy in enumerate(xys):
                                # if there are multiple branches, we append a suffix to the name
                                suffix = f"_{i}" if len(xys) > 1 else ""
                                outflow_reporters[
                                    f"river_outflow_hourly_m3_per_s_{river_ID}{suffix}"
                                ] = {
                                    "varname": "grid.var.discharge_m3_s_per_substep",
                                    "type": "grid",
                                    "function": f"sample_xy,{xy[0]},{xy[1]}",
                                    "substeps": 24,
                                }
                        report_config = multi_level_merge(
                            report_config,
                            {"hydrology.routing": outflow_reporters},
                        )
                    elif module_name == "_water_circle":
                        if module_values is True:
                            report_config = multi_level_merge(
                                report_config,
                                WATER_CIRCLE_REPORT_CONFIG,
                            )
                    elif module_name == "_water_balance":
                        if module_values is True:
                            report_config = multi_level_merge(
                                report_config,
                                WATER_BALANCE_REPORT_CONFIG,
                            )
                    elif module_name == "_energy_balance":
                        if module_values is True:
                            report_config = multi_level_merge(
                                report_config,
                                ENERGY_BALANCE_REPORT_CONFIG,
                            )
                    else:
                        raise ValueError(
                            f"Module {module_name} is not a valid module for reporting."
                        )

                    to_delete.append(module_name)

            for module_name in to_delete:
                del self.model.config["report"][module_name]

            for module_name, configs in self.model.config["report"].items():
                self.variables[module_name] = {}
                for name, config in configs.items():
                    assert isinstance(config, dict), (
                        f"Configuration for {module_name}.{name} must be a dictionary, but is {type(config)}."
                    )
                    self.variables[module_name][name] = self.create_variable(
                        config, module_name, name
                    )
        else:
            self.is_activated = False

    def create_variable(self, config: dict, module_name: str, name: str) -> list | None:
        """This function creates a variable for the reporter.

        For
        configurations without a aggregation function, a zarr file is created. For
        configurations with an aggregation function, the data
        is stored in memory and exported on the final timestep.

        Args:
            config: The configuration for the variable.
            module_name: The name of the module to which the variable belongs.
            name: The name of the variable.

        Returns:
            A list of values if the variable is scalar, None otherwise.

        Raises:
            ValueError: If the variable type is not recognized.
        """
        if config["type"] == "scalar":
            assert "function" not in config or config["function"] is None, (
                "Scalar variables cannot have a function. "
            )
            return []
        elif config["type"] in ("grid", "HRU"):
            if config["function"] is not None:
                return []
            else:
                return None

        elif config["type"] == "agents":
            if config["function"] is not None:
                return []
            else:
                return None

        else:
            raise ValueError(
                f"Type {config['type']} not recognized. Must be 'scalar', 'grid', 'agents' or 'HRU'."
            )

    def maybe_report_value(
        self,
        module_name: str,
        name: str,
        module: Any,
        local_variables: dict,
        config: dict,
    ) -> None:
        """This method is used to save and/or export model values.

        Args:
            module_name: Name of the module to which the value belongs.
            name: Name of the value to be exported.
            module: The module to report data from.
            local_variables: A dictionary of local variables from the function
                that calls this one.
            config: Configuration for saving the file. Contains options such a file format, and whether to export the data or save the data in the model.

        Raises:
            ValueError: If the frequency is not recognized.
            KeyError: If the variable is not found in the local variables or module attributes.
            AttributeError: If the attribute is not found in the module.
        """
        # here we return None if the value is not to be reported on this timestep
        if "frequency" in config:
            if config["frequency"] == "initial":
                if self.model.current_timestep != 0:
                    return None
            elif config["frequency"] == "final":
                if self.model.current_timestep != self.model.n_timesteps - 1:
                    return None
            elif "every" in config["frequency"]:
                every = config["frequency"]["every"]
                if every == "year":
                    month = config["frequency"]["month"]
                    day = config["frequency"]["day"]
                    if (
                        self.model.current_time.month != month
                        or self.model.current_time.day != day
                    ):
                        return None
                elif every == "month":
                    day = config["frequency"]["day"]
                    if self.model.current_time.day != day:
                        return None
                elif every == "day":
                    pass
                else:
                    raise ValueError(
                        f"Frequency every {config['every']} not recognized (must be 'yearly', or 'monthly')."
                    )
            else:
                raise ValueError(f"Frequency {config['frequency']} not recognized.")

        varname = config["varname"]
        fancy_index = re.search(r"\[.*?\]", varname)
        if fancy_index:
            fancy_index = fancy_index.group(0)
            varname = varname.replace(fancy_index, "")

        # get the variable
        if varname.startswith("."):
            varname = varname[1:]
            try:
                value = local_variables[varname]
            except KeyError:
                raise KeyError(
                    f"Variable {varname} not found in local variables of {module_name}. "
                )
        else:
            try:
                value = attrgetter(varname)(module)
            except AttributeError:
                raise AttributeError(
                    f"Attribute {varname} not found in module {module_name}. "
                )

        if fancy_index:
            value = eval(f"value{fancy_index}")

        # if the value is not None, we check whether the value is valid
        if isinstance(value, list):
            value = np.array([v.item() for v in value])
            for v in value:
                assert not np.isnan(value) and not np.isinf(v)
        elif np.isscalar(value):
            assert not np.isnan(value) and not np.isinf(value)
            if isinstance(value, (np.floating, np.integer, np.bool_)):
                value = value.item()

        self.process_value(module_name, name, value, config)

    def process_value(
        self,
        module_name: str,
        name: str,
        value: np.ndarray | int | float | bool,
        config: dict,
    ) -> None:
        """Exports an array of values to the export folder.

        Args:
            module_name: Name of the module to which the value belongs.
            name: Name of the value to be exported.
            value: The array itself.
            config: Configuration for saving the file. Contains options such a file format, and whether to export the array in this timestep at all.

        Raises:
            ValueError: If the function is not recognized,
                or if the variable type is not recognized.
            IndexError: If the coordinate is in sample_lon_lat is outside the model domain.
        """
        type_: str | None = config.get("type", None)
        if type_ is None:
            raise ValueError(f"Type not specified for {config}.")

        if "function" in config and config["function"] is None and "path" not in config:
            zarr_path: Path = self.report_folder / module_name / (name + ".zarr")
            zarr_path.parent.mkdir(parents=True, exist_ok=True)
            config["path"] = str(zarr_path)

            store = zarr.storage.LocalStore(zarr_path, read_only=False)
            config["_store"] = store

            root_group = zarr.open_group(store, mode="w")
            config["_root_group"] = root_group

            config["_index"] = 0

        if type_ in ("grid", "HRU"):
            if not isinstance(value, np.ndarray):
                raise ValueError(
                    f"Value for {module_name}.{name} must be a numpy ndarray, but is {type(value)}."
                )

            grid_size = self.hydrology.grid.compressed_size
            HRU_size = self.hydrology.HRU.compressed_size
            if type_ == "grid" and value.shape[-1] != grid_size:
                error = f"Value for {module_name}.{name} has size {value.shape[-1]}, but grid size is {grid_size}."
                if value.shape[-1] == HRU_size:
                    error += f" HRU is size {HRU_size}. Did you mean to set type to 'HRU' instead of 'grid'?"
                raise ValueError(error)

            if type_ == "HRU" and value.shape[-1] != HRU_size:
                error = f"Value for {module_name}.{name} has size {value.shape[-1]}, but HRU size is {HRU_size}."
                if value.shape[-1] == grid_size:
                    error += f" grid id size {grid_size}. Did you mean to set type to 'grid' instead of 'HRU'?"
                raise ValueError(error)

            # in case of no aggregation function, we write the data directly to zarr
            if config["function"] is None:
                if type_ == "HRU":
                    value: np.ndarray = self.hydrology.HRU.decompress(value)
                else:
                    value: np.ndarray = self.hydrology.grid.decompress(value)

                # in the first timestep, we create the array that will hold the actual data
                if value.ndim == 3:
                    substeps: int = value.shape[0]
                else:
                    substeps: int = 1

                if config["_index"] == 0:  # first time writing data
                    if config["type"] == "HRU":
                        raster = self.hydrology.HRU
                    else:
                        raster = self.hydrology.grid

                    time = create_time_array(
                        start=self.model.simulation_start,
                        end=self.model.simulation_end,
                        timestep=self.model.timestep_length,
                        conf=config,
                        substeps=substeps,
                    )

                    chunk_size: int = get_time_chunk_size(
                        value.dtype,
                        raster.lat.size,
                        raster.lon.size,
                        target_size_bytes=self.config["chunk_target_size_bytes"],
                    )
                    chunk_size: int = min(chunk_size, time.size)

                    # ensure chunk size is multiple of substeps
                    chunk_size: int = max(chunk_size // substeps, 1) * substeps

                    assert isinstance(raster.crs, str)
                    prepare_gridded_group(
                        name,
                        config,
                        raster.lon,
                        raster.lat,
                        raster.crs,
                        value,
                        time,
                        chunk_size,
                        self.config["compression_level"],
                    )

                root_group = config["_root_group"]

                buffer = config["_chunk_data"]
                chunk_size: int = buffer.shape[0]

                # Calculate the index in the buffer
                start_index = config["_index"] % chunk_size
                end_index = start_index + substeps
                buffer[start_index:end_index, ...] = value

                # If the buffer is full, flush it to disk
                if end_index == chunk_size:
                    chunk_index = config["_index"] // chunk_size
                    self._flush_chunk_data(root_group, name, buffer, chunk_index)

                config["_index"] += substeps
                return
            else:
                # in the first timestep, we create the array that will hold the actual data
                if value.ndim == 2:
                    substeps: int = value.shape[0]
                    config["substeps"] = substeps
                else:
                    substeps: int = 1

                function, *args = config["function"].split(",")
                if function == "mean":
                    value = np.mean(value, axis=-1)
                elif function == "nanmean":
                    value = np.nanmean(value, axis=-1)
                elif function == "sum":
                    value = np.sum(value, axis=-1)
                elif function == "nansum":
                    value = np.nansum(value, axis=-1)
                elif function in ("sample_xy", "sample_lonlat"):
                    # for sample_xy, args are pixel coordinates, which we
                    # first need to convert to their pixel index in x and y
                    if function == "sample_lonlat":
                        if type_ == "grid":
                            gt: tuple[float, float, float, float, float, float] = (
                                self.model.hydrology.grid.gt
                            )
                        elif type_ == "HRU":
                            gt: tuple[float, float, float, float, float, float] = (
                                self.hydrology.HRU.gt
                            )
                        else:
                            raise ValueError(
                                f"Unknown varname type {config['varname']}"
                            )
                        px, py = coord_to_pixel((float(args[0]), float(args[1])), gt)
                    else:
                        # for sample_xy, args are pixel indices
                        px: int = int(args[0])
                        py: int = int(args[1])

                    # both for the grid and HRU, the data is stored efficiently, so that
                    # all data is in a 1D array, and we need to map the pixel coordinates
                    # to the index in that 1D array
                    # therefore, we first get the linear mapping and use that to get the index
                    # then we extract the value at that index
                    if type_ == "grid":
                        linear_mapping: TwoDArrayInt32 = (
                            self.hydrology.grid.linear_mapping
                        )
                    elif type_ == "HRU":
                        linear_mapping: TwoDArrayInt32 = (
                            self.hydrology.HRU.linear_mapping
                        )
                    else:
                        raise ValueError(f"Unknown varname type {config['varname']}")

                    try:
                        idx = linear_mapping[py, px]
                    except IndexError:
                        raise IndexError(
                            f"Coordinate ({px}, {py}) is outside the model domain, which has shape {linear_mapping.shape}."
                        )
                    if idx == -1:
                        raise IndexError(
                            f"Coordinate ({px}, {py}) is not a valid cell."
                        )

                    # extract the value at that index
                    value = value[..., idx]
                elif function in (
                    "weightedmean",
                    "weightednanmean",
                    "weightedsum",
                    "weightednansum",
                ):
                    if type_ == "HRU":
                        cell_area = self.hydrology.HRU.var.cell_area
                    else:
                        cell_area = self.hydrology.grid.var.cell_area
                    if function == "weightedmean":
                        value = np.average(value, weights=cell_area, axis=-1)
                    elif function == "weightednanmean":
                        value = np.nansum(value * cell_area, axis=-1) / np.sum(
                            cell_area
                        )
                    elif function == "weightedsum":
                        value = np.sum(value * cell_area, axis=-1)
                    elif function == "weightednansum":
                        value = np.nansum(value * cell_area, axis=-1)

                else:
                    raise ValueError(f"Function {function} not recognized")

        elif type_ == "agents":
            if config["function"] is None:
                if config["_index"] == 0:
                    time = create_time_array(
                        start=self.model.simulation_start,
                        end=self.model.simulation_end,
                        timestep=self.model.timestep_length,
                        conf=config,
                    )

                    prepare_agent_group(
                        name,
                        config,
                        time,
                        value,
                        self.config["chunk_target_size_bytes"],
                        self.config["compression_level"],
                    )

                root_group = config["_root_group"]
                value_store_array = root_group[name]
                assert isinstance(value_store_array, zarr.Array)

                assert isinstance(value, (np.ndarray, DynamicArray))
                if value.size < value_store_array.shape[0]:
                    print("Padding array with NaNs or -1 - temporary solution")
                    value = np.pad(
                        value.data if isinstance(value, DynamicArray) else value,
                        (0, value_store_array.shape[0] - value.size),
                        mode="constant",
                        constant_values=get_fill_value(value)[0] or False,
                    )

                buffer = config["_chunk_data"]
                chunk_size: int = buffer.shape[0]

                # Calculate the index in the buffer
                index = config["_index"] % chunk_size
                buffer[index, ...] = value

                # If the buffer is full, flush it to disk
                if index + 1 == chunk_size:
                    chunk_index = config["_index"] // chunk_size
                    self._flush_chunk_data(root_group, name, buffer, chunk_index)

                config["_index"] += 1

                return
            else:
                function, *args = config["function"].split(",")
                if function == "mean":
                    value = np.mean(value)
                elif function == "nanmean":
                    value = np.nanmean(value)
                elif function == "sum":
                    value = np.sum(value)
                elif function == "nansum":
                    value = np.nansum(value)
                else:
                    raise ValueError(f"Function {function} not recognized")

        if module_name not in self.variables:
            self.variables[module_name] = {}

        if name not in self.variables[module_name]:
            self.variables[module_name][name] = []

        if "substeps" in config:
            assert isinstance(value, np.ndarray)
            assert len(value) == config["substeps"], (
                f"Value for {module_name}.{name} has length {len(value)}, but {config['substeps']} substeps are expected."
            )
            self.variables[module_name][name].extend(
                [
                    (
                        self.model.current_time
                        + i * self.model.timestep_length / config["substeps"],
                        v,
                    )
                    for i, v in enumerate(value)
                ]
            )
        else:
            self.variables[module_name][name].append((self.model.current_time, value))

    def _flush_chunk_data(
        self, group: zarr.Group, name: str, buffer: np.ndarray, chunk_index: int
    ) -> None:
        """Flush a specific chunk for grid/HRU variables.

        We write directly to the zarr blocks to avoid the overhead of zarr's
        indexing and slicing. This is possible because we are writing full chunks.

        Args:
            group: The zarr group where the variable is stored.
            name: The name of the variable.
            buffer: The buffer containing the data to flush.
            chunk_index: The index of the chunk to flush.
        """
        zarr_array = group[name]
        assert isinstance(zarr_array, zarr.Array)
        zarr_array.blocks[chunk_index, ...] = buffer

    def finalize(self) -> None:
        """At the end of the model run, all previously collected data is reported to disk."""
        # If no data has been collected, we return
        if self.model.config["report"] is None:
            self.model.logger.info("No report configuration found. No data to report.")
            return

        # Flush any remaining buffers
        for module_name, configs in self.model.config["report"].items():
            for name, config in configs.items():
                if "function" in config and config["function"] is None:
                    chunk_time_size: int = config["_chunk_data"].shape[0]
                    buffer_end: int = config["_index"] % chunk_time_size
                    if buffer_end == 0:
                        continue  # nothing to flush
                    chunk_index: int = config["_index"] // chunk_time_size
                    self._flush_chunk_data(
                        group=config["_root_group"],
                        name=name,
                        buffer=config["_chunk_data"][:buffer_end, ...],
                        chunk_index=chunk_index,
                    )

        # Export all scalar and aggregated variables to CSV
        for module_name, variables in self.variables.items():
            for name, values in variables.items():
                if self.model.config["report"][module_name][name][
                    "type"
                ] == "scalar" or (
                    self.model.config["report"][module_name][name]["function"]
                    is not None
                ):
                    # if the variable is a scalar or has an aggregation function, we report
                    df = pd.DataFrame.from_records(
                        values, columns=["time", name], index="time"
                    )

                    folder = self.report_folder / module_name
                    folder.mkdir(parents=True, exist_ok=True)

                    df.to_csv(folder / (name + ".csv"))

                    fig, ax = plt.subplots(figsize=(30, 5))
                    fig.tight_layout()

                    df.plot(y=name, title=f"{module_name}.{name}", ax=ax)
                    plt.grid()
                    plt.savefig(folder / (name + ".svg"), format="svg")
                    plt.close()

    def report(
        self, module: Module, local_variables: dict[str, Any], module_name: str
    ) -> None:
        """This method is in every step function to report data to disk.

        Args:
            module: The module to report data from.
            local_variables: A dictionary of local variables from the function
                that calls this one.
            module_name: The name of the module.
        """
        if not self.is_activated:
            return None
        report = self.model.config["report"].get(module_name, None)
        if report is not None:
            for name, config in report.items():
                self.maybe_report_value(
                    module_name=module_name,
                    name=name,
                    module=module,
                    local_variables=local_variables,
                    config=config,
                )

    @property
    def is_activated(self) -> bool:
        """Returns whether the reporter is activated.

        Returns:
            True if the reporter is activated, False otherwise.
        """
        return self._is_activated

    @is_activated.setter
    def is_activated(self, value: bool) -> None:
        self._is_activated = value



"""This module contains the Industry agent class for simulating industrial water demand in the GEB model."""

from __future__ import annotations

import calendar
import datetime
from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from geb.geb_types import ArrayFloat32
from geb.hydrology.HRUs import load_water_demand_xr
from geb.store import Bucket

from ..hydrology.landcovers import SEALED
from .general import AgentBaseClass, downscale_volume

if TYPE_CHECKING:
    from geb.agents import Agents
    from geb.model import GEBModel


class IndustryVariables(Bucket):
    """Variables for the LivestockFarmers agent."""

    current_water_demand: ArrayFloat32
    current_efficiency: ArrayFloat32
    last_water_demand_update: datetime.datetime


class Industry(AgentBaseClass):
    """This class is used to simulate industry.

    Note:
        Currently, this module is not actually agent-based but rather
        uses aggregated pre-defined water demand data.

    Args:
        model: The GEB model.
        agents: The class that includes all agent types (allowing easier communication between agents).
    """

    var: IndustryVariables

    def __init__(self, model: GEBModel, agents: Agents) -> None:
        """Initialize the Industry agent module.

        Args:
            model: The GEB model.
            agents: The class that includes all agent types (allowing easier communication between agents).
        """
        super().__init__(model)

        if self.model.simulate_hydrology:
            self.HRU = model.hydrology.HRU
            self.grid = model.hydrology.grid

        self.agents = agents
        self.config = (
            self.model.config["agent_settings"]["industry"]
            if "industry" in self.model.config["agent_settings"]
            else {}
        )

        self.industry_water_consumption_ds: xr.Dataset = load_water_demand_xr(
            self.model.files["other"]["water_demand/industry_water_consumption"]
        )
        self.industry_water_demand_ds: xr.Dataset = load_water_demand_xr(
            self.model.files["other"]["water_demand/industry_water_demand"]
        )
        if self.model.in_spinup:
            self.spinup()

    @property
    def name(self) -> str:
        """The name of the module.

        Used to save data to disk.

        Returns:
            The name of the module.
        """
        return "agents.industry"

    def spinup(self) -> None:
        """Set initial water demand and efficiency during spinup."""
        water_demand, efficiency = self.update_water_demand()
        self.var.current_water_demand = water_demand
        self.var.current_efficiency = efficiency

    def update_water_demand(self) -> tuple[ArrayFloat32, ArrayFloat32]:
        """Update the water demand for industry at the HRU level.

        Returns:
            A tuple containing:
            - The updated water demand (in m3/day).
            - The updated efficiency [0-1].
        """
        if self.config.get("disable_water_demand", False):
            self.model.logger.info(
                "[Industry] Water demand and efficiency set to 0 due to config setting."
            )
            zero_array = np.zeros(self.HRU.var.land_use_ratio.shape, dtype=float)
            efficiency_grid = self.model.hydrology.to_grid(
                HRU_data=zero_array, fn="max"
            )
            self.var.last_water_demand_update = self.model.current_time
            return zero_array, efficiency_grid

        else:
            downscale_mask = self.model.hydrology.HRU.var.land_use_type != SEALED
            days_in_year = 366 if calendar.isleap(self.model.current_time.year) else 365

            water_demand = (
                self.industry_water_demand_ds.sel(
                    time=self.model.current_time,
                    method="ffill",
                    tolerance="366D",  # ty:ignore[invalid-argument-type]
                ).industry_water_demand
                * 1_000_000
                / days_in_year
            )
            water_demand = (
                water_demand.rio.write_crs(4326).rio.reproject(
                    4326,
                    shape=self.grid.shape,
                    transform=self.grid.transform,
                )
                / (water_demand.rio.transform().a / self.grid.transform.a) ** 2
            )  # correct for change in cell size
            water_demand = (
                downscale_volume(
                    water_demand.rio.transform().to_gdal(),
                    self.grid.gt,
                    water_demand.values,
                    self.grid.mask,
                    self.model.hydrology.mapping_grid_to_HRU_uncompressed,
                    downscale_mask,
                    self.HRU.var.land_use_ratio,
                )
                / self.HRU.var.cell_area
            )  # convert to m/day

            water_demand = (
                water_demand
                * self.model.config["parameters"]["water_demand_multiplier_industry"]
            )

            water_consumption = (
                self.industry_water_consumption_ds.sel(
                    time=self.model.current_time, method="ffill"
                ).industry_water_consumption
                * 1_000_000
                / days_in_year
            )
            water_consumption = (
                water_consumption.rio.write_crs(4326).rio.reproject(
                    4326,
                    shape=self.grid.shape,
                    transform=self.grid.transform,
                )
                / (water_consumption.rio.transform().a / self.grid.transform.a) ** 2
            )
            water_consumption = (
                downscale_volume(
                    water_consumption.rio.transform().to_gdal(),
                    self.grid.gt,
                    water_consumption.values,
                    self.grid.mask,
                    self.model.hydrology.mapping_grid_to_HRU_uncompressed,
                    downscale_mask,
                    self.HRU.var.land_use_ratio,
                )
                / self.HRU.var.cell_area
            )  # convert to m/day

            efficiency = np.divide(
                water_consumption,
                water_demand,
                out=np.zeros_like(water_consumption, dtype=float),
                where=water_demand != 0,
            )

            efficiency = self.model.hydrology.to_grid(HRU_data=efficiency, fn="max")

            assert (efficiency <= 1).all()
            assert (efficiency >= 0).all()
            self.var.last_water_demand_update = self.model.current_time
            return water_demand, efficiency

    def water_demand(self) -> tuple[ArrayFloat32, ArrayFloat32]:
        """Get the current water demand for industry at the HRU level.

        Updates the water demand only if data for this timestep is available.
        Otherwise, assumes the last known water demand.

        Returns:
            A tuple containing:
            - The current water demand (in m3/day).
            - The current efficiency [0-1].
        """
        if (
            np.datetime64(self.model.current_time, "ns")
            in self.industry_water_consumption_ds.time
        ):
            water_demand, efficiency = self.update_water_demand()
            self.var.current_water_demand = water_demand
            self.var.current_efficiency = efficiency

        assert (
            self.model.current_time - self.var.last_water_demand_update
        ).days < 366, (
            "Water demand has not been updated for over a year. "
            "Please check the industry water demand datasets."
        )
        return self.var.current_water_demand, self.var.current_efficiency

    def step(self) -> None:
        """This function is run each timestep."""
        self.report(locals())



"""Module to handle climate forcing data."""

from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Literal, overload

import numpy as np
import numpy.typing as npt
import xarray as xr

from geb.geb_types import ArrayFloat32, ArrayFloat64, ThreeDArrayFloat32
from geb.workflows.io import read_grid, read_table

from .module import Module
from .workflows.io import AsyncGriddedForcingReader

if TYPE_CHECKING:
    from geb.model import GEBModel


def generate_bilinear_interpolation_weights(
    src_x: ArrayFloat64,
    src_y: ArrayFloat64,
    tgt_x: ArrayFloat64,
    tgt_y: ArrayFloat64,
) -> tuple[ArrayFloat32, ArrayFloat32]:
    """
    Generates indices and weights for bilinear interpolation.

    Assumes the source grid is rectilinear and monotonic.

    Args:
        src_x: Source grid x-coordinates (must be monotonic).
        src_y: Source grid y-coordinates (must be monotonic).
        tgt_x: Target points x-coordinates.
        tgt_y: Target points y-coordinates.

    Returns:
        tuple: A tuple containing:
            - indices (ndarray): Shape (n_targets, 4) of flat source grid indices.
            - weights (ndarray): Shape (n_targets, 4) of interpolation weights.

    Raises:
        ValueError: If target points are outside the source grid bounds.
    """
    # Create the full grid of target points and flatten them
    # The 'xy' indexing creates a target grid of shape (len(tgt_y), len(tgt_x))
    # Note: len(tgt_y) is the row count (Ny_tgt), len(tgt_x) is the col count (Nx_tgt)
    tgt_x_2d, tgt_y_2d = np.meshgrid(tgt_x, tgt_y, indexing="xy")

    # Flatten the target coordinates into (N_targets,) arrays
    tgt_x = tgt_x_2d.flatten()
    tgt_y = tgt_y_2d.flatten()

    # Source Grid Dimensions
    nx: int = len(src_x)
    ny: int = len(src_y)

    # Validate that x-coordinates are ascending
    if not np.all(src_x[1:] > src_x[:-1]):
        raise ValueError("Source x-coordinates must be strictly ascending.")

    # ix is the index of the source x-coordinate that is just less than or equal to tgt_x
    ix = np.searchsorted(src_x, tgt_x) - 1
    # Clip indices to be within valid bounds for x
    ix = np.clip(ix, 0, nx - 2)
    # Check bounds for ix after clipping - if any targets are still out of bounds, it's an error
    if not np.all((tgt_x >= src_x[0]) & (tgt_x <= src_x[-1])):
        raise ValueError("Target x-coordinates are outside the source grid bounds.")

    is_y_descending = src_y[0] > src_y[-1]

    if not is_y_descending:
        # For ascending y-axis (normal case)
        iy = np.searchsorted(src_y, tgt_y, side="right") - 1
        iy = np.clip(iy, 0, ny - 2)
        y0 = src_y[iy]
        y1 = src_y[iy + 1]
        dy = (tgt_y - y0) / (y1 - y0)
    else:
        # For descending y-axis, use vectorized NumPy operations
        # Create a matrix comparison to find intervals efficiently
        # src_y[j] >= tgt_y >= src_y[j+1] for descending coordinates

        # Reshape for broadcasting: tgt_y as column, src_y as row
        tgt_y_col = tgt_y[:, np.newaxis]  # Shape: (n_targets, 1)
        src_y_intervals = src_y[:-1]  # Shape: (ny-1,) - upper bounds of intervals
        src_y_intervals_next = src_y[1:]  # Shape: (ny-1,) - lower bounds of intervals

        # Find where target points fall within each interval
        # For descending: src_y[j] >= target_y >= src_y[j+1]
        in_interval = (src_y_intervals >= tgt_y_col) & (
            tgt_y_col >= src_y_intervals_next
        )

        # Find the first (leftmost) interval index for each target point
        iy = np.argmax(in_interval, axis=1)

        # Handle out-of-bounds cases
        no_interval_found = ~np.any(in_interval, axis=1)
        above_highest = tgt_y > src_y[0]
        below_lowest = tgt_y < src_y[-1]

        # Assign boundary indices for out-of-bounds points
        iy = np.where(no_interval_found & above_highest, 0, iy)
        iy = np.where(no_interval_found & below_lowest, ny - 2, iy)

        # Calculate dy for descending coordinates
        y0 = src_y[iy]
        y1 = src_y[iy + 1]
        # For descending coordinates, the weight should be calculated as:
        # dy = (y0 - target_y) / (y0 - y1) since y0 > y1
        dy = (y0 - tgt_y) / (y0 - y1)  # Get the corner coordinates
    x0 = src_x[ix]
    x1 = src_x[ix + 1]

    # Calculate normalized distances
    dx = (tgt_x - x0) / (x1 - x0)

    # Weights for the four corners
    w00 = (1 - dx) * (1 - dy)  # Corresponds to (y0, x0)
    w01 = dx * (1 - dy)  # Corresponds to (y0, x1)
    w10 = (1 - dx) * dy  # Corresponds to (y1, x0)
    w11 = dx * dy  # Corresponds to (y1, x1)

    # Flat indices for the four corners
    idx00 = iy * nx + ix
    idx01 = iy * nx + (ix + 1)
    idx10 = (iy + 1) * nx + ix
    idx11 = (iy + 1) * nx + (ix + 1)

    # Stack indices and weights in a consistent order
    indices = np.stack([idx00, idx01, idx10, idx11], axis=1).astype(np.int32)
    weights = np.stack([w00, w01, w10, w11], axis=1).astype(np.float32)

    assert (weights >= 0).all() and (weights <= 1).all(), "Weights must be in [0, 1]"
    assert np.allclose(weights.sum(axis=1), 1.0), "Weights must sum to 1"

    return indices, weights


def get_pressure_correction_factor(
    DEM: npt.NDArray[np.float32],
    g: float,
    Mo: float,
    lapse_rate: float,
) -> npt.NDArray[np.float32]:
    """Calculate pressure correction factor based on elevation.

    Args:
        DEM: Digital elevation model data (meters).
        g: Gravitational constant (m/s²).
        Mo: Molecular weight of dry air (kg/mol).
        lapse_rate: Temperature lapse rate (K/m).

    Returns:
        Pressure correction factor (dimensionless).
    """
    return (288.15 / (288.15 + lapse_rate * DEM)) ** (g * Mo / (8.3144621 * lapse_rate))


class ForcingLoader(ABC):
    """Abstract base class for loading and validating forcing data."""

    def __init__(
        self, model: GEBModel, variable: str, n: int, supports_forecast: bool = True
    ) -> None:
        """Initialize the ForcingLoader.

        Args:
            model: The GEB model instance.
            variable: The variable name to load (e.g., "pr" for precipitation).
            n: Number of time steps to load at once (default is 1).
            supports_forecast: Whether the loader supports forecast mode.
        """
        self.model: GEBModel = model
        self.n: int = n
        self.variable: str = variable
        self.reader: AsyncGriddedForcingReader = AsyncGriddedForcingReader(
            model.files["other"][f"climate/{variable}"], variable, asynchronous=True
        )

        self.indices, self.weights = generate_bilinear_interpolation_weights(
            self.reader.x,
            self.reader.y,
            model.hydrology.grid.lon,
            model.hydrology.grid.lat,
        )
        self.xsize: int = model.hydrology.grid.lon.size
        self.ysize: int = model.hydrology.grid.lat.size

        self._supports_forecast = supports_forecast

        self._in_forecast_mode: bool = False
        self.ds_forecast: xr.DataArray | None = None
        self.forecast_issue_datetime: datetime | None = None

    @property
    def in_forecast_mode(self) -> bool:
        """Indicates whether the loader is in forecast mode.

        Returns:
            True if in forecast mode, False otherwise.
        """
        return self._in_forecast_mode

    @in_forecast_mode.setter
    def in_forecast_mode(self, value: bool) -> None:
        self._in_forecast_mode = value

    @property
    def supports_forecast(self) -> bool:
        """Indicates whether the loader supports forecast mode.

        Returns:
            True if forecast mode is supported, False otherwise.
        """
        return self._supports_forecast

    def load(self, dt: datetime) -> ThreeDArrayFloat32:
        """Load and validate forcing data for a given time.

        If in forecast mode and the time is after the forecast issue date,
        the data is loaded from the forecast dataset.

        Otherwise, data is loaded from the standard reader.

        Args:
            dt: The datetime for which to load the data.

        Returns:
            The interpolated and validated data as a numpy array.

        Raises:
            ValueError: If the data is invalid according to the validation criteria.
        """
        # check if we are in forecasting mode, and if the end of the timestep is after the
        # start of the forecast
        if (
            self.forecast_issue_datetime is not None
            and dt + self.model.timestep_length >= self.forecast_issue_datetime
        ):
            if self.ds_forecast is None:
                raise ValueError(
                    "Forecast data array is not set, but loader is in forecast mode."
                )
            # find how many substeps to load from the normal data source
            substeps_to_forecast: int = int(
                (self.forecast_issue_datetime - dt).total_seconds()
                / (self.model.timestep_length / self.n).total_seconds()
            )
            # if some substeps are before the forecast, load them from the normal reader
            if substeps_to_forecast > 0:
                # TODO: The reader breaks when loading less than n timesteps, so we load n and slice
                # the data
                non_forecast_data_all, normal_start_date = self.reader.read_timestep(
                    dt, n=self.n
                )

                if normal_start_date != np.datetime64(dt, "ns"):
                    raise ValueError(
                        f"Standard reader returned data starting at {normal_start_date}, but expected {dt}."
                    )

                non_forecast_data: npt.NDArray[np.float32] = non_forecast_data_all[
                    :substeps_to_forecast, :, :
                ]

                forecast_data: npt.NDArray[np.float32] = self.ds_forecast.isel(
                    time=slice(None, self.n - substeps_to_forecast)
                ).values
                data: npt.NDArray[np.float32] = np.concatenate(
                    [non_forecast_data, forecast_data], axis=0
                )
            else:
                # if the current day is already fully in forecast, load all from forecast
                data: npt.NDArray[np.float32] = self.ds_forecast.sel(
                    time=slice(
                        dt,
                        dt
                        + self.model.timestep_length
                        - self.model.timestep_length / self.n,
                    )
                ).values
        else:
            data_all, normal_start_date = self.reader.read_timestep(dt, n=self.n)
            if normal_start_date != np.datetime64(dt, "ns"):
                raise ValueError(
                    f"Standard reader returned data starting at {normal_start_date}, but expected {dt}."
                )
            data: npt.NDArray[np.float32] = data_all

        interpolated: npt.NDArray[np.float32] = self.interpolate(data)
        valid: bool = self.validate(interpolated)
        if not valid:
            raise ValueError(
                f"Invalid data for time {dt} for variable {self.variable}."
            )
        assert interpolated.dtype == np.dtype(np.float32)
        return interpolated

    def set_forecast(self, forecast_issue_datetime: datetime, da: xr.DataArray) -> None:
        """Sets the loader to forecast mode.

        This means that data from before the forecast issue date will be loaded
        normally, while data from the forecast issue date onwards will be loaded
        from the forecast data file. Data beyond the available forecast data will
        raise an error.

        Args:
            forecast_issue_datetime: The datetime when the forecast starts.
            da: The xarray DataArray containing the forecast data.
        """
        self.forecast_issue_datetime: datetime = forecast_issue_datetime
        self.ds_forecast: xr.DataArray = da

    def unset_forecast(self) -> None:
        """Unset forecast mode."""
        self.ds_forecast: None = None
        self.forecast_issue_datetime: None = None

    @property
    def in_forecast_mode(self) -> bool:
        """Indicates whether the loader is in forecast mode.

        Returns:
            True if in forecast mode, False otherwise.
        """
        return self.forecast_issue_datetime is not None

    def interpolate(self, data: npt.NDArray[np.float32]) -> npt.NDArray[Any]:
        """Interpolate data to the model grid using bilinear interpolation.

        Args:
            data: The input data array to interpolate.

        Returns:
            The interpolated data array.
        """
        data_flattened_xy_dims = data.reshape(data.shape[0], -1)
        # the corner values must be gathered in the same order as the weights
        corner_values = data_flattened_xy_dims[:, self.indices]
        interpolated_flattened_xy_dims = np.sum(
            corner_values * self.weights[np.newaxis, :, :], axis=2
        )
        output = interpolated_flattened_xy_dims.reshape(
            interpolated_flattened_xy_dims.shape[0], self.ysize, self.xsize
        )
        return output

    def validate(self, v: npt.NDArray[np.float32]) -> bool:
        """Validate the data array.

        Args:
            v: The data array to validate.

        Returns:
            True, indicating that by default all data is valid.

        Raises:
            ValueError: If the data shape does not match the expected grid shape.
        """
        if v.shape[0] != self.n:
            raise ValueError(f"Data time dimension does not match expected n {self.n}.")
        if v.shape[1] != self.ysize or v.shape[2] != self.xsize:
            raise ValueError(
                f"Data shape {v.shape[1:]} does not match expected grid shape "
                f"({self.ysize}, {self.xsize})."
            )
        if not self.validate_values(v):
            raise ValueError("Data validation failed.")

        return True

    @abstractmethod
    def validate_values(self, v: npt.NDArray[np.float32]) -> bool:
        """Validate the data array.

        Args:
            v: The data array to validate.

        Returns:
            True if valid, otherwise raises ValueError.
        """
        pass


class Precipitation(ForcingLoader):
    """Loader for precipitation data with specific validation."""

    def __init__(self, model: GEBModel, grid_mask: npt.NDArray[np.bool_]) -> None:
        """Initialize the Precipitation loader.

        Args:
            model: The GEB model instance.
            grid_mask: The mask of the model grid (True for valid points, False for invalid).
        """
        self.grid_mask = grid_mask
        super().__init__(model, "pr_kg_per_m2_per_s", 24)

    def validate_values(self, v: npt.NDArray[np.float32]) -> bool:
        """Validate precipitation data.

        Args:
            v: The data array to validate.

        Returns:
            True if valid, otherwise raises ValueError.
        """
        v_non_masked = v[:, self.grid_mask]
        return ((v_non_masked >= 0).all() and (v_non_masked < 500 / 3600).all()).item()


class Temperature(ForcingLoader):
    """Loader for temperature data with specific validation."""

    def __init__(
        self,
        model: GEBModel,
        forcing_DEM: npt.NDArray[np.float32],
        grid_DEM: npt.NDArray[np.float32],
        grid_mask: npt.NDArray[np.bool_],
        dewpoint: bool = False,
        lapse_rate: float = -0.0065,
    ) -> None:
        """Initialize the Temperature loader.

        This class performs a lapse rate correction based on the difference between the
        forcing DEM and the model grid DEM.

        Args:
            model: The GEB model instance.
            forcing_DEM: The DEM used in the forcing data.
            grid_DEM: The DEM used in the model grid.
            grid_mask: The mask of the model grid (True for valid points, False for invalid).
            dewpoint: If True, load dewpoint temperature; otherwise, load air temperature.
            lapse_rate: The lapse rate in K/m (default is -0.0065 K/m).
        """
        self.forcing_DEM = forcing_DEM
        self.grid_DEM = grid_DEM
        self.grid_mask = grid_mask
        self.lapse_rate = lapse_rate

        if dewpoint:
            super().__init__(model, "dewpoint_tas_2m_K", 24)
        else:
            super().__init__(model, "tas_2m_K", 24)

    def validate_values(self, v: npt.NDArray[np.float32]) -> bool:
        """Validate temperature data.

        Args:
            v: The data array to validate.

        Returns:
            True if valid, otherwise raises ValueError.
        """
        v_non_masked = v[:, self.grid_mask]
        return ((v_non_masked > 170).all() and (v_non_masked < 370).all()).item()

    def interpolate(self, data: npt.NDArray[np.float32]) -> npt.NDArray[Any]:
        """Interpolate data to the model grid using bilinear interpolation.

        Overrides the base class method to handle temperature with a lapse rate correction.
        First, the temperature is adjusted to sea level using the forcing DEM and lapse rate.
        Then, after interpolation, it is adjusted back to the model grid elevation.

        This ensures a smooth temperature gradient with elevation.

        Args:
            data: The input data array to interpolate.

        Returns:
            The interpolated data array.
        """
        temperature_sea_level = (
            data - self.forcing_DEM[np.newaxis, :, :] * self.lapse_rate
        )
        interpolated_temperature_sea_level = super().interpolate(data)
        interpolated_temperature = (
            interpolated_temperature_sea_level
            + self.grid_DEM[np.newaxis, :, :] * self.lapse_rate
        )

        return interpolated_temperature


class Wind(ForcingLoader):
    """Loader for wind data with specific validation.

    Note that wind can be both positive and negative.
    """

    def __init__(
        self, model: GEBModel, direction: str, grid_mask: npt.NDArray[np.bool_]
    ) -> None:
        """Initialize the Wind loader.

        Args:
            model: The GEB model instance.
            direction: The wind direction, either "u" or "v".
            grid_mask: The mask of the model grid (True for valid points, False for invalid).

        Raises:
            ValueError: If the direction is not "u" or "v".
        """
        if direction not in ["u", "v"]:
            raise ValueError("Direction must be 'u' or 'v'")
        self.grid_mask = grid_mask
        super().__init__(model, f"wind_{direction}10m_m_per_s", 24)

    def validate_values(self, v: npt.NDArray[np.float32]) -> bool:
        """Validate wind data.

        Args:
            v: The data array to validate.

        Returns:
            True if valid, otherwise raises ValueError.
        """
        v_non_masked = v[:, self.grid_mask]
        return ((v_non_masked >= -150).all() and (v_non_masked < 150).all()).item()


class Pressure(ForcingLoader):
    """Loader for surface pressure data with elevation-based correction and validation."""

    def __init__(
        self,
        model: GEBModel,
        forcing_DEM: npt.NDArray[np.float32],
        grid_DEM: npt.NDArray[np.float32],
        grid_mask: npt.NDArray[np.bool_],
        g: float = 9.80665,
        Mo: float = 0.0289644,
        lapse_rate: float = -0.0065,
    ) -> None:
        """Initialize the Pressure loader.

        This class performs an elevation-based pressure correction based on the difference
        between the forcing DEM and the model grid DEM.

        Args:
            model: The GEB model instance.
            forcing_DEM: The DEM used in the forcing data (meters).
            grid_DEM: The DEM used in the model grid (meters).
            grid_mask: The mask of the model grid (True for valid points, False for invalid).
            g: Gravitational constant (m/s², default is 9.80665).
            Mo: Molecular weight of dry air (kg/mol, default is 0.0289644).
            lapse_rate: Temperature lapse rate (K/m, default is -0.0065).
        """
        self.forcing_DEM = forcing_DEM
        self.grid_DEM = grid_DEM
        self.grid_mask = grid_mask
        self.g = g
        self.Mo = Mo
        self.lapse_rate = lapse_rate

        super().__init__(model, "ps_pascal", 24)

    def validate_values(self, v: npt.NDArray[np.float32]) -> bool:
        """Validate surface pressure data.

        Args:
            v: The data array to validate (Pa).

        Returns:
            True if valid, otherwise raises ValueError.
        """
        v_non_masked = v[:, self.grid_mask]
        return ((v_non_masked > 30_000).all() and (v_non_masked < 120_000).all()).item()

    def interpolate(self, data: npt.NDArray[np.float32]) -> npt.NDArray[Any]:
        """Interpolate data to the model grid using bilinear interpolation.

        Overrides the base class method to handle pressure with elevation correction.
        First, the pressure is adjusted to sea level using the forcing DEM and atmospheric model.
        Then, after interpolation, it is adjusted back to the model grid elevation.

        This ensures proper pressure gradients with elevation.

        Args:
            data: The input pressure data array to interpolate (Pa).

        Returns:
            The interpolated pressure data array (Pa).
        """
        # Convert pressure to sea level by dividing by correction factor
        # for the forcing grid
        pressure_sea_level = (
            data
            / get_pressure_correction_factor(
                self.forcing_DEM, self.g, self.Mo, self.lapse_rate
            )[np.newaxis, :, :]
        )

        # Interpolate the sea level pressure
        interpolated_pressure_sea_level = super().interpolate(pressure_sea_level)

        # Convert back to grid elevation by multiplying by correction factor, now
        # for the model grid
        interpolated_pressure = (
            interpolated_pressure_sea_level
            * get_pressure_correction_factor(
                self.grid_DEM, self.g, self.Mo, self.lapse_rate
            )[np.newaxis, :, :]
        )

        return interpolated_pressure


class RSDS(ForcingLoader):
    """Loader for surface downwelling shortwave radiation data with specific validation."""

    def __init__(self, model: GEBModel, grid_mask: npt.NDArray[np.bool_]) -> None:
        """Initialize the RSDS loader.

        Args:
            model: The GEB model instance.
            grid_mask: The mask of the model grid (True for valid points, False for invalid).
        """
        self.grid_mask = grid_mask
        super().__init__(model, "rsds_W_per_m2", 24)

    def validate_values(self, v: npt.NDArray[np.float32]) -> bool:
        """Validate surface downwelling shortwave radiation data.

        Args:
            v: The data array to validate.

        Returns:
            True if valid, otherwise raises ValueError.
        """
        v_non_masked = v[:, self.grid_mask]
        return (v_non_masked >= 0).all().item()


class RLDS(ForcingLoader):
    """Loader for surface downwelling longwave radiation data with specific validation."""

    def __init__(self, model: GEBModel, grid_mask: npt.NDArray[np.bool_]) -> None:
        """Initialize the RLDS loader.

        Args:
            model: The GEB model instance.
            grid_mask: The mask of the model grid (True for valid points, False for invalid).
        """
        self.grid_mask = grid_mask
        super().__init__(model, "rlds_W_per_m2", 24)

    def validate_values(self, v: npt.NDArray[np.float32]) -> bool:
        """Validate surface downwelling longwave radiation data.

        Args:
            v: The data array to validate.

        Returns:
            True if valid, otherwise raises ValueError.
        """
        v_non_masked = v[:, self.grid_mask]
        return (v_non_masked >= 0).all().item()


class SPEI(ForcingLoader):
    """Loader for Standardized Precipitation-Evapotranspiration Index (SPEI) data with specific validation."""

    def __init__(self, model: GEBModel, grid_mask: npt.NDArray[np.bool_]) -> None:
        """Initialize the SPEI loader.

        Args:
            model: The GEB model instance.
            grid_mask: The mask of the model grid (True for valid points, False for invalid).
        """
        self.grid_mask = grid_mask
        super().__init__(model, "SPEI", 1, supports_forecast=False)

    def validate_values(self, v: npt.NDArray[np.float32]) -> bool:
        """Validate SPEI data.

        Args:
            v: The data array to validate.

        Returns:
            True if valid, otherwise raises ValueError.
        """
        v_non_masked = v[:, self.grid_mask]
        return not np.isnan(v_non_masked).any()


class CO2:
    """Loader for CO2 concentration data with specific validation."""

    def __init__(self, model: GEBModel) -> None:
        """Initialize the CO2 loader."""
        self.df = read_table(model.files["table"]["climate/CO2_ppm"])

    def load(self, time: datetime) -> float:
        """Load CO2 concentration data for a given time.

        Args:
            time: The datetime for which to load the data.

        Returns:
            The CO2 concentration value.

        Raises:
            ValueError: If the data is invalid according to the validation criteria.
        """
        data: float = self.df.loc[time.year].item()
        valid: bool = self.validate_values(data)
        if not valid:
            raise ValueError(f"Invalid CO2 data for time {time}.")
        return data

    def validate_values(self, v: float) -> bool:
        """Validate CO2 concentration data.

        Args:
            v: the data value to validate.

        Returns:
            True if valid, otherwise raises ValueError.
        """
        return v > 270 and v < 2000

    @property
    def supports_forecast(self) -> bool:
        """Indicates whether the loader supports forecast mode.

        Returns:
            False as CO2 loader does not support forecast mode.
        """
        return False

    def set_forecast(self, forecast_issue_datetime: datetime, da: xr.DataArray) -> None:
        """Set forecast mode.

        CO2 loader does not support forecast mode.

        Raises:
            NotImplementedError: As CO2 loader does not support forecast mode.
        """
        raise NotImplementedError("CO2 loader does not support forecast mode.")

    def unset_forecast(self) -> None:
        """Unset forecast mode.

        CO2 loader does not support forecast mode.

        Raises:
            NotImplementedError: As CO2 loader does not support forecast mode.
        """
        raise NotImplementedError("CO2 loader does not support forecast mode.")


class Forcing(Module):
    """Module to handle climate forcing data.

    This module is responsible for loading and validating climate forcing data such as temperature, humidity, pressure, and radiation.
    It provides methods to load specific datasets and ensures that the data meets certain validation criteria.

    Args:
        model: The GEB model instance.
    """

    def __init__(self, model: GEBModel) -> None:
        """Initialize the Forcing module.

        All forcing loaders are initialized upfront to allow for efficient batch loading
        and inter-variable dependencies during interpolation.

        Args:
            model: The GEB model instance.
        """
        self.model = model
        self.forcing_DEM = read_grid(model.files["other"]["climate/elevation_forcing"])

        # Initialize all forcing loaders upfront
        self._initialize_loaders()

    def _initialize_loaders(self) -> None:
        """Initialize all forcing loaders.

        This creates instances of all forcing loaders so they're ready to use.
        """
        grid_mask = ~self.model.hydrology.grid.mask
        grid_DEM = self.model.hydrology.grid.decompress(
            self.model.hydrology.grid.var.elevation
        )

        # Initialize all loaders
        self._loaders: dict[str, CO2 | ForcingLoader] = {
            "pr_kg_per_m2_per_s": Precipitation(self.model, grid_mask=grid_mask),
            "tas_2m_K": Temperature(
                self.model,
                forcing_DEM=self.forcing_DEM,
                grid_DEM=grid_DEM,
                grid_mask=grid_mask,
                dewpoint=False,
            ),
            "dewpoint_tas_2m_K": Temperature(
                self.model,
                forcing_DEM=self.forcing_DEM,
                grid_DEM=grid_DEM,
                grid_mask=grid_mask,
                dewpoint=True,
            ),
            "wind_u10m_m_per_s": Wind(self.model, direction="u", grid_mask=grid_mask),
            "wind_v10m_m_per_s": Wind(self.model, direction="v", grid_mask=grid_mask),
            "ps_pascal": Pressure(
                self.model,
                forcing_DEM=self.forcing_DEM,
                grid_DEM=grid_DEM,
                grid_mask=grid_mask,
            ),
            "rsds_W_per_m2": RSDS(self.model, grid_mask=grid_mask),
            "rlds_W_per_m2": RLDS(self.model, grid_mask=grid_mask),
            "SPEI": SPEI(self.model, grid_mask=grid_mask),
            "CO2_ppm": CO2(self.model),
        }

    @property
    def name(self) -> str:
        """Return the name of the module.

        Returns:
            The name of the module.
        """
        return "forcing"

    @overload
    def __getitem__(self, name: Literal["CO2_ppm"]) -> CO2: ...

    @overload
    def __getitem__(self, name: str) -> ForcingLoader: ...

    def __getitem__(self, name: str) -> ForcingLoader | CO2:
        """Get the forcing loader for a given name.

        Args:
            name: name of forcing dataset, e.g. "pr_kg_per_m2_per_s", "tas_2m_K", etc.

        Returns:
            The forcing loader for the specified dataset.

        Raises:
            KeyError: If the forcing variable name is not recognized.
        """
        if name not in self._loaders:
            raise KeyError(
                f"Forcing variable '{name}' not found. Available variables: {list(self._loaders.keys())}"
            )
        return self._loaders[name]

    @overload
    def load(self, name: Literal["CO2_ppm"], dt: datetime | None = None) -> float: ...

    @overload
    def load(self, name: str, dt: datetime | None = None) -> ThreeDArrayFloat32: ...

    def load(self, name: str, dt: datetime | None = None) -> ThreeDArrayFloat32 | float:
        """Load forcing data for a given name and time.

        Args:
            name: name of forcing dataset, e.g. "pr_kg_per_m2_per_s", "tas_2m_K", etc.
            dt: time of forcing data to be returned. Defaults to None, in which case
                the current time of the model is used.

        Returns:
            Forcing data as a numpy array or float.
        """
        if dt is None:
            # the forcing data is the accumulated data over the current hour.
            # Therefore, we load the data for the end of the hour to ensure we get
            # the accumulated value for the day.
            # For instantaneous values like temperature, we load the data for the
            # start with the first full hour of the day. Perhaps it would
            # be very slightly better to take the midpoint of the hour, but this
            # would make the code more complex and the difference should be negligible,
            # so we take the start of the hour for simplicity.
            dt: datetime = self.model.current_time + timedelta(hours=1)

        return self[name].load(dt)

    @property
    def loaders(self) -> dict[str, CO2 | ForcingLoader]:
        """Get all forcing loaders.

        Returns:
            A dictionary of all forcing loaders.
        """
        return self._loaders

    @property
    def forcing_loaders(self) -> dict[str, ForcingLoader]:
        """Get all forcing loaders except CO2.

        Returns:
            A dictionary of all forcing loaders except CO2.
        """
        return {k: v for k, v in self._loaders.items() if isinstance(v, ForcingLoader)}

    def spinup(self) -> None:
        """Prepare the forcing module for the simulation.

        Does not do anything as no spinup is needed.
        """
        pass

    def step(self) -> None:
        """Advance the forcing module by one time step.

        Does not do anything as the forcing data is read on demand.
        """
        pass



"""File adapter for datasets stored as files.

Useful for simple files that only need to be downloaded and stored without additional processing.
"""

from __future__ import annotations

from typing import Any

from geb.workflows.io import fetch_and_save

from .base import Adapter


class File(Adapter):
    """Adapter for File datasets."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the File adapter."""
        super().__init__(*args, **kwargs)

    def fetch(self, url: str) -> File:
        """Fetch the dataset from the given URL if not already present.

        Args:
            url: The URL to fetch the dataset from.

        Returns:
            The GlobGM adapter instance.
        """
        if not self.is_ready:
            fetch_and_save(url=url, file_path=self.path)
        return self



"""Utility functions for flood hazard workflows."""

import os
import platform
import shutil
import subprocess
from datetime import datetime
from os.path import isfile, join
from pathlib import Path
from typing import Any

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
import xarray as xr
from hydromt_sfincs import SfincsModel, utils
from matplotlib.cm import viridis  # ty: ignore[unresolved-import]
from shapely import line_locate_point
from shapely.geometry import GeometryCollection, LineString, Point

from geb.geb_types import ArrayFloat32, ArrayInt64
from geb.workflows.io import read_geom


def export_rivers(
    model_root: Path, rivers: gpd.GeoDataFrame, postfix: str = ""
) -> None:
    """Export river segments to a GeoParquet file.

    Args:
        model_root: The root path of the model directory.
        rivers: The GeoDataFrame containing the river segments.
        postfix: A postfix to add to the filename. Defaults to "".
    """
    rivers.to_parquet(model_root / f"rivers{postfix}.geoparquet")


def import_rivers(model_root: Path, postfix: str = "") -> gpd.GeoDataFrame:
    """Import river segments from a GeoParquet file.

    Args:
        model_root: The root path of the model directory.
        postfix: A postfix to add to the filename. Defaults to "".

    Returns:
        A GeoDataFrame containing the river segments.
    """
    return read_geom(model_root / f"rivers{postfix}.geoparquet")


def read_flood_depth(
    model_root: Path,
    simulation_root: Path,
    method: str,
    minimum_flood_depth: float,
    end_time: datetime,
    mask: xr.DataArray | None = None,
) -> xr.DataArray:
    """Read the maximum flood depth from the SFINCS model results.

    If SFINCS was run with subgrid, the flood depth is downscaled to subgrid resolution.

    Notes:
        If SFINCS was run as a fluvial model, a minimum flood depth of 0.15 m is applied.
        If SFINCS was run as a pluvial/coastal model, a minimum flood depth of 0.05 m is applied.
        There are some fundamental issues with the current generation of subgrid maps,
            especially in steep terrain, see conclusion here: https://doi.org/10.5194/gmd-18-843-2025
            Here, we use bilinear interpolation as a goodish solution. However, we keep our
            eyes out for better solutions.

    Args:
        model_root: The root path of the SFINCS model directory. This should contain the subgrid directory with the depth file.
        simulation_root: The root path of the SFINCS simulation directory. This should contain the simulation results files.
        minimum_flood_depth: The minimum flood depth to apply during downscaling (in meters).
        method: The method to use for calculating flood depth. Options are 'max' for maximum flood depth
            and 'final' for flood depth at the final time step.
        end_time: The end time of the simulation.
        mask: Optional xarray DataArray mask to apply to the flood depth. Defaults to None.

    Returns:
        The maximum flood depth downscaled to subgrid resolution.

    Raises:
        ValueError: If an unknown method is provided.
        KeyError: If the specified end_time is not in the results time dimension.
    """
    # Read SFINCS model, config and results
    model: SfincsModel = SfincsModel(
        str(simulation_root),
        mode="r",
    )
    model.read_config()

    # For unknown reasons, sometimes reading the results fails the first time
    # but succeeds the second time. Therefore, we try twice here.
    try:
        model.read_results()
    except OSError:
        model.read_results()

    # to detect whether SFINCS was run with subgrid, we check if the 'sbgfile' key exists in the config
    # to be extra safe, we also check if the value is not None or has has length > 0
    if (
        "sbgfile" in model.config
        and model.config["sbgfile"] is not None
        and len(model.config["sbgfile"]) > 0
    ):
        if method == "max":
            # get maximum water surface elevation (with respect to sea level)
            water_surface_elevation = model.results["zsmax"].max(dim="timemax")
            assert isinstance(water_surface_elevation, xr.DataArray)
        elif method == "final":
            # get water surface elevation at the final time step (with respect to sea level)
            all_water_surface_elevation: xr.Dataset | xr.DataArray = model.results["zs"]
            assert isinstance(all_water_surface_elevation, xr.DataArray)
            try:
                water_surface_elevation: xr.DataArray = all_water_surface_elevation.sel(
                    time=end_time
                )
            except KeyError:
                raise KeyError(
                    f"The specified end_time {end_time} is not in the results time dimension. \
                    Consider whether the reporting interval of SFINCS is expected \
                    to include the end_time."
                )
            assert isinstance(water_surface_elevation, xr.DataArray)
        else:
            raise ValueError(f"Unknown method: {method}")

        if mask is not None:
            mask = mask.compute()
            water_surface_elevation = water_surface_elevation.compute()
            water_surface_elevation = water_surface_elevation.where(mask, np.nan)
        # read subgrid elevation
        surface_elevation: xr.DataArray = (
            xr.open_dataarray(model_root / "subgrid" / "dep_subgrid.tif")
            .sel(band=1)
            .drop_vars(["band"])
        )

        flood_depth_m: xr.DataArray = utils.downscale_floodmap(
            zsmax=water_surface_elevation,
            dep=surface_elevation,
            hmin=minimum_flood_depth,
            reproj_method="bilinear",  # maybe use "nearest" for coastal
        )
        flood_depth_m: xr.DataArray = flood_depth_m.rio.write_crs(model.crs)

    else:
        if method == "max":
            flood_depth_m: xr.Dataset | xr.DataArray = model.results["hmax"].max(
                dim="timemax"
            )
            assert isinstance(flood_depth_m, xr.DataArray)
        elif method == "final":
            flood_depth_m_all_steps: xr.Dataset | xr.DataArray = model.results["h"]
            assert isinstance(flood_depth_m_all_steps, xr.DataArray)
            try:
                flood_depth_m: xr.DataArray = flood_depth_m_all_steps.sel(time=end_time)
            except KeyError:
                raise KeyError(
                    f"The specified end_time {end_time} is not in the results time dimension. \
                    Consider whether the reporting interval of SFINCS is expected \
                    to include the end_time."
                )

        else:
            raise ValueError(f"Unknown method: {method}")

        flood_depth_m: xr.DataArray = xr.where(
            flood_depth_m >= minimum_flood_depth, flood_depth_m, np.nan, keep_attrs=True
        )
        if mask is not None:
            flood_depth_m = flood_depth_m.where(mask.values, np.nan)

        flood_depth_m.attrs["_FillValue"] = np.nan

    print(
        f"Maximum flood depth: {float(flood_depth_m.max().values):.2f} m, "
        f"Mean flood depth: {float(flood_depth_m.mean().values):.2f} m"
    )
    # Create basemap plot
    fig, ax = model.plot_basemap(
        fn_out=None,  # ty: ignore[invalid-argument-type]
        variable="",  # No variable to plot, only basemap
        plot_geoms=False,
        zoomlevel=12,
        figsize=(11, 7),  # ty: ignore[invalid-argument-type]
    )

    # Plot flood depth with colorbar
    cbar_kwargs: dict[str, float | tuple[int, int]] = {"shrink": 0.6, "anchor": (0, 0)}
    flood_depth_m.plot(
        x="x",
        y="y",
        ax=ax,
        vmin=0,
        vmax=float(flood_depth_m.max().values),
        cmap=viridis,
        cbar_kwargs=cbar_kwargs,
    )

    ax.set_title("Maximum Water Depth over all time steps")

    output_path: Path = model_root / "flood_depth_all_time_steps.png"
    plt.savefig(output_path, dpi=300, bbox_inches="tight")
    return flood_depth_m


def to_sfincs_datetime(dt: datetime) -> str:
    """Convert a datetime object to a string in the format required by SFINCS.

    Args:
        dt: datetime object to convert.

    Returns:
        String representation of the datetime in the format "YYYYMMDD HHMMSS".
    """
    return dt.strftime("%Y%m%d %H%M%S")


def make_relative_paths(
    config: dict[str, Any],
    model_root: Path,
    new_root: Path,
) -> dict[str, Any]:
    """Return dict with paths to the new model new root.

    Always use POSIX format for paths, to ensure compatibility with Docker.

    Args:
        config: The configuration dictionary with paths to make relative.
        model_root: The original model root directory.
        new_root: The new model root directory.

    Returns:
        A dictionary with the updated paths relative to the new model root directory.

    Raises:
        ValueError: If model_root and new_root do not have a common path.
    """
    commonpath = ""
    if os.path.splitdrive(model_root)[0] == os.path.splitdrive(new_root)[0]:
        commonpath = os.path.commonpath([model_root, new_root])
    if os.path.basename(commonpath) == "":
        raise ValueError("model_root and new_root must have a common path")
    relpath = Path(os.path.relpath(commonpath, new_root))

    config_kwargs = dict()
    for k, v in config.items():
        if (
            isinstance(v, str)
            and isfile(join(model_root, v))
            and not isfile((join(new_root, v)))
        ):
            # Ensure paths are consistent across platforms
            path = Path(relpath) / Path(v)  # Ensure both are Path objects
            config_kwargs[k] = (
                path.as_posix()
            )  # Always use POSIX format for Docker compatibility

    return config_kwargs


def run_sfincs_subprocess(
    working_directory: Path, cmd: list[str], log_file: Path
) -> int:
    """Runs SFINCS model in a subprocess.

    Args:
        working_directory: The working directory for the subprocess.
        cmd: The command to run as a list of strings.
        log_file: The file to write the log output to.

    Returns:
        The return code of the subprocess.

    """
    print(f"Running SFINCS with: {cmd}")
    with open(file=log_file, mode="w") as log:
        with subprocess.Popen(
            args=cmd,
            cwd=working_directory,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        ) as process:
            stdout = process.stdout
            stderr = process.stderr
            assert stdout is not None
            assert stderr is not None

            # Continuously read lines from stdout and stderr
            for line in iter(lambda: stdout.readline() or stderr.readline(), ""):
                print(line.rstrip())
                log.write(line)
                log.flush()

            process.wait()

    return process.returncode


def get_start_point(geom: LineString) -> Point:
    """Extract the start point from a LineString geometry.

    Args:
        geom: LineString geometry.

    Returns:
        The start point as a Shapely Point object.
    """
    return Point(geom.coords[0])


def get_end_point(geom: LineString) -> Point:
    """Extract the end point from a LineString geometry.

    Args:
        geom: LineString geometry.

    Returns:
        The end point as a Shapely Point object.
    """
    return Point(geom.coords[-1])


def create_hourly_hydrograph(
    peak_discharge: float,
    rising_limb_hours: int,
    recession_limb_hours: int,
) -> pd.DataFrame:
    """Create a triangular hydrograph time series.

    Args:
        peak_discharge: The peak discharge of the hydrograph.
        rising_limb_hours: The duration of the rising limb in hours.
        recession_limb_hours: The duration of the recession limb in hours.

    Returns:
        A pandas DataFrame with the time series of the hydrograph.
    """
    total_duration_hours: int | float = rising_limb_hours + recession_limb_hours
    # Create a time array with hourly intervals
    time_hours: ArrayInt64 = np.arange(0, total_duration_hours + 1)
    # Create discharge array for triangular shape
    discharge: ArrayFloat32 = np.zeros_like(time_hours, dtype=np.float32)
    # Rising limb: linear increase to peak
    discharge[: rising_limb_hours + 1] = np.linspace(
        0, peak_discharge, rising_limb_hours + 1
    )
    # Recession limb: linear decrease from peak
    discharge[rising_limb_hours:] = np.linspace(
        peak_discharge, 0, recession_limb_hours + 1
    )
    # Create a pandas DataFrame for the time series
    time_index: pd.DatetimeIndex = pd.date_range(
        start="2024-01-01 00:00", periods=len(time_hours), freq="h"
    )
    hydrograph_df: pd.DataFrame = pd.DataFrame(
        {"time": time_index, "discharge": discharge}
    )
    hydrograph_df = hydrograph_df.set_index("time")
    return hydrograph_df


def check_docker_running() -> bool | None:
    """Check if Docker is installed and running.

    Returns:
        True if Docker is installed and running, False otherwise.

    """
    try:
        subprocess.run(
            ["docker", "info"],
            check=True,
            capture_output=True,
        )
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        print("Docker not installed properly or not running properly.")
        return False


def run_sfincs_simulation(
    model_root: Path,
    simulation_root: Path,
    ncpus: int | str = "auto",
    gpu: bool | str = "auto",
) -> int:
    """Run SFINCS simulation using either Apptainer or Docker.

    Args:
        model_root: Path to the model root directory.
        simulation_root: Path to the simulation root directory.
            Some paths in the configuration will be made relative to this path.
            The simulation directory must be a subdirectory of the model root directory.
        ncpus: Number of CPUs to use. Can be an integer or 'auto' to automatically detect the number of CPUs.
        gpu: Whether to use GPU support. Can be True, False, or 'auto'. In auto mode,
            the presence of an NVIDIA GPU is checked using `nvidia-smi`. Defaults to auto.

    Raises:
        ValueError: If gpu is not True, False, or 'auto'.
        RuntimeError: If there is an error running the SFINCS simulation.

    Returns:
        The return code of the SFINCS simulation subprocess.
    """
    if gpu not in [True, False, "auto"]:
        raise ValueError("gpu must be True, False, or 'auto'")

    if gpu == "auto":
        # check if nvidia-smi is available, if so, check if a GPU is present
        if shutil.which(cmd="nvidia-smi") is not None:
            result: subprocess.CompletedProcess[bytes] = subprocess.run(
                args=["nvidia-smi"], capture_output=True
            )
            maybe_gpu: bool = result.returncode == 0

            # if no GPU is found, there is no need to check further, set gpu to False
            if not maybe_gpu:
                gpu: bool = False
            else:
                # otherwise, check if we are in a SLURM job. We may be on a cluster
                # where a GPU is physically present, but not allocated to the job
                # and therefore not available for use.
                in_SLURM_job: bool = "SLURM_JOB_ID" in os.environ
                if in_SLURM_job:
                    # in a SLURM job, if the job has access to the GPU
                    gpu_ids: str | None = os.getenv(key="CUDA_VISIBLE_DEVICES")
                    if gpu_ids is not None and len(gpu_ids) > 0:
                        gpu: bool = True
                    else:
                        gpu: bool = False
                else:
                    gpu: bool = True
        else:
            gpu: bool = False

        if gpu:
            print("GPU detected, running SFINCS with GPU support.")
        else:
            print("No GPU detected, running SFINCS without GPU support.")

    if gpu:
        version: str | None = os.getenv(key="SFINCS_CONTAINER_GPU")
        assert version is not None, (
            "SFINCS_CONTAINER_GPU environment variable is not set"
        )
    else:
        version: str | None = os.getenv(key="SFINCS_CONTAINER")
        assert version is not None, "SFINCS_CONTAINER environment variable is not set"

    if platform.system() == "Linux":
        # If not a apptainer image, add docker:// prefix
        # to the version string
        if not version.endswith(".sif"):
            version: str = "docker://" + version

        c = (
            int(
                os.getenv("SLURM_CPUS_PER_TASK", None)
                or os.getenv("SLURM_CPUS_ON_NODE", None)
                or os.cpu_count()  # returns none if cannot be determined
                or 1
            )
            if ncpus == "auto"
            else int(ncpus)
        )
        ncpus_str = "0" if c == 1 else f"0-{c - 1}"

        cmd: list[str] = [
            "taskset",
            "-c",
            ncpus_str,  # get user defined or automatically detected number of CPUs
            "apptainer",
            "run",
            "-B",  ## Bind mount
            f"{model_root.resolve()}:/data",
            "--pwd",  ## Set working directory inside container
            f"/data/{simulation_root.relative_to(model_root)}",
            "--nv",
            version,
        ]

    else:
        assert check_docker_running(), "Docker is not running"
        # in the docker image of SFINCS thet working directory is /data
        # but since we also want to load files that are from the model
        # root, we mount the model root to /data and then change the
        # working directory within the Docker image to the simulation root
        cmd: list[str] = [
            "docker",
            "run",
            "-v",
            f"{model_root.resolve()}:/data",
            "-w",
            f"/data/{simulation_root.relative_to(model_root).as_posix()}",
            version,
        ]

    log_file: Path = simulation_root / "sfincs.log"
    return_code: int = run_sfincs_subprocess(simulation_root, cmd, log_file=log_file)
    if return_code != 0:
        raise RuntimeError(
            f"Error running SFINCS simulation: {return_code}. The used command was: {' '.join(cmd)}. The log file is located at {log_file}."
        )
    else:
        return return_code


def _get_xy(
    river: pd.Series,
    up_to_downstream: bool = True,
) -> tuple[int, int]:
    """Get the first valid xy coordinate from a river's hydrography_xy list.

    Starts upstream or downstream based on the up_to_downstream flag.

    Args:
        river: Series containing river information, including 'hydrography_xy'.
            This is a list of (x, y) tuples, representing the location of the river
            in the low-resolution hydrological grid.
        up_to_downstream: Whether to search from upstream to downstream.
            Defaults to True (starting upstream).

    Returns:
        A tuple (x, y) of the first valid coordinate.
    """
    xys = river["hydrography_xy"]
    if up_to_downstream:
        idx: int = 0
    else:
        idx: int = -1
    return (xys[idx][0], xys[idx][1])


def get_representative_river_points(
    river_ID: set, rivers: pd.DataFrame
) -> list[tuple[int, int]]:
    """Get representative river points for a given river ID.

    If the river is represented in the low-resolution hydrological grid, its
    representative point is used. If not, the function traverses upstream to find
    rivers that are represented in the grid and uses their representative points.

    Args:
        river_ID: The ID of the river for which to find representative points.
        rivers: DataFrame containing river information, including 'represented_in_grid' and 'hydrography_xy'.

    Returns:
        A list of tuples (x, y) representing the coordinates of the representative points.
        If no valid points are found, an empty list is returned.

    Raises:
        ValueError: If no valid xy coordinates are found for rivers.
    """
    river = rivers.loc[river_ID]
    if river["represented_in_grid"]:
        xy = _get_xy(river, up_to_downstream=True)
        if xy is not None:
            return [xy]
        else:
            raise ValueError(
                f"Error: No valid xy found for river {river_ID} which is represented in the grid."
            )

    else:
        river_IDs = set([river_ID])
        representitative_rivers = set()
        while river_IDs:
            river_ID = river_IDs.pop()
            river = rivers.loc[river_ID]
            if not river["represented_in_grid"]:
                upstream_rivers = rivers[rivers["downstream_ID"] == river_ID]
                river_IDs.update(upstream_rivers.index)
            else:
                representitative_rivers.add(river_ID)

        representitative_rivers = rivers[rivers.index.isin(representitative_rivers)]
        xys = []
        for river_ID, river in representitative_rivers.iterrows():
            xy = _get_xy(river, up_to_downstream=False)
            if xy is not None:
                xys.append(xy)
            else:
                raise ValueError(
                    f"Error: No valid xy found for river {river_ID} which is represented not in the grid. Likely because upstream rivers could not be found."
                )

        return xys


def get_discharge_and_river_parameters_by_river(
    river_IDs: list[int],
    points_per_river: list[list[tuple[int, int]]],
    discharge: xr.DataArray,
    river_width_alpha: npt.NDArray[np.float32] | None = None,
    river_width_beta: npt.NDArray[np.float32] | None = None,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Extract discharge time series and river parameters for each river.

    When rivers are represented in the low-resolution hydrological grid, rivers
    should have only one point in the points_per_river list. When rivers are not
    represented in the low-resolution hydrological grid, the river is represented
    by its upstream rivers, which can be multiple. In that case, the discharge
    time series for the river is the sum of the discharge time series of all its
    representative points.

    Args:
        river_IDs: List of river IDs.
        points_per_river: List of lists of (x, y) tuples representing the points for each river.
        discharge: xarray DataArray containing discharge values with dimensions (time, y, x).
        river_width_alpha: 2D array of river width alpha parameters, same shape as discharge y and x dimensions.
            If None, river width alpha will not be extracted. Defaults to None.
        river_width_beta: 2D array of river width beta parameters, same shape as discharge y and x dimensions.
            If None, river width beta will not be extracted. Defaults to None.

    Returns:
        A tuple containing:
            - A pandas DataFrame with discharge time series for each river (columns are river IDs).
            - A pandas DataFrame with river parameters (index is river IDs, columns are 'river_width_alpha' and 'river_width_beta').

    Raises:
        ValueError: If no points are found for rivers or if discharge values contain NaNs.
    """
    xs: list[int] = []
    ys: list[int] = []
    for points in points_per_river:
        xs.extend([p[0] for p in points])
        ys.extend([p[1] for p in points])

    if len(xs) == 0 or len(ys) == 0:
        raise ValueError("No points found for rivers.")

    x_points: xr.DataArray = xr.DataArray(
        xs,
        dims="points",
    )
    y_points: xr.DataArray = xr.DataArray(
        ys,
        dims="points",
    )

    discharge_per_point: xr.DataArray = discharge.isel(
        x=x_points,
        y=y_points,
    ).compute()
    assert not np.isnan(discharge_per_point.values).any(), (
        "Discharge values contain NaNs"
    )

    if river_width_alpha is not None:
        river_width_alpha_per_point = river_width_alpha[y_points, x_points]
    else:
        river_width_alpha_per_point = None
    if river_width_beta is not None:
        river_width_beta_per_point = river_width_beta[y_points, x_points]
    else:
        river_width_beta_per_point = None

    discharge_df: pd.DataFrame = pd.DataFrame(index=discharge.time)
    discharge_df.index.freq = "h"  # ty:ignore[invalid-assignment]
    river_parameters: pd.DataFrame = pd.DataFrame(
        index=np.array(river_IDs),
        columns=np.array(
            ["river_width_alpha", "river_width_beta"],
        ),
    )

    i: int = 0
    for river_ID, points in zip(river_IDs, points_per_river, strict=True):
        discharge_per_river = discharge_per_point.isel(
            points=slice(i, i + len(points))
        ).sum(dim="points")

        discharge_df[river_ID] = discharge_per_river

        if river_width_alpha_per_point is not None:
            river_width_alpha_per_river = river_width_alpha_per_point[
                i : i + len(points)
            ].mean()
            if np.isnan(river_width_alpha_per_river):
                print(
                    f"Warning: River width alpha for river {river_ID} is NaN. Setting to 0."
                )
                river_width_alpha_per_river = 0.0
            river_parameters.loc[river_ID, "river_width_alpha"] = (
                river_width_alpha_per_river
            )

        if river_width_beta_per_point is not None:
            river_width_beta_per_river = river_width_beta_per_point[
                i : i + len(points)
            ].mean()
            if np.isnan(river_width_beta_per_river):
                print(
                    f"Warning: River width beta for river {river_ID} is NaN. Setting to 1."
                )
                river_width_beta_per_river = 1.0
            river_parameters.loc[river_ID, "river_width_beta"] = (
                river_width_beta_per_river
            )

        i += len(points)

    assert i == len(xs), "Discharge values do not match the number of points"
    assert i == len(ys), "Discharge values do not match the number of points"
    # make sure no NaN values are present in the discharge DataFrame
    assert not discharge_df.isnull().values.any(), "Discharge DataFrame contains NaNs"

    if river_width_alpha_per_point is not None:
        # make sure no NaN values are present in the river parameters DataFrame
        assert not river_parameters["river_width_alpha"].isnull().values.any(), (
            "River width alpha DataFrame contains NaNs"
        )
    if river_width_beta_per_point is not None:
        # make sure no NaN values are present in the river parameters DataFrame
        assert not river_parameters["river_width_beta"].isnull().values.any(), (
            "River width beta DataFrame contains NaNs"
        )
    return discharge_df, river_parameters


def select_most_upstream_point(
    river: LineString, outflow_points: GeometryCollection
) -> Point:
    """Select the most upstream point from a collection of outflow points.

    Args:
        river: LineString of the river geometry.
        outflow_points: GeometryCollection of outflow points (can contain Points and LineStrings).

    Returns:
        The most upstream Point from the outflow_points.

    Raises:
        TypeError: If an unsupported geometry type is found in outflow_points.
    """
    points: list[Point] = []
    for geom in outflow_points.geoms:
        if isinstance(geom, Point):
            points.append(geom)
        elif isinstance(geom, LineString):
            for coord in geom.coords:
                points.append(Point(coord))
        else:
            raise TypeError(
                f"Unsupported geometry type in outflow_points: {type(geom)}"
            )

    most_upstream_point: Point = points[0]
    most_upstream_point_loc: float = line_locate_point(river, most_upstream_point)
    for point in points[1:]:
        loc = line_locate_point(river, point)
        if loc < most_upstream_point_loc:
            most_upstream_point = point
            most_upstream_point_loc = loc

    outflow_point: Point = most_upstream_point

    return outflow_point



"""Typing definitions for GEB."""

from typing import Any, TypeVar

import numpy as np
from numpy.typing import DTypeLike

Shape = TypeVar("Shape", bound=tuple[int, ...])


ArrayFloat16 = np.ndarray[tuple[int], np.dtype[np.float16]]
ArrayFloat32 = np.ndarray[tuple[int], np.dtype[np.float32]]
ArrayFloat64 = np.ndarray[tuple[int], np.dtype[np.float64]]
ArrayFloat = ArrayFloat16 | ArrayFloat32 | ArrayFloat64

ArrayInt8 = np.ndarray[tuple[int], np.dtype[np.int8]]
ArrayUint8 = np.ndarray[tuple[int], np.dtype[np.uint8]]
ArrayInt16 = np.ndarray[tuple[int], np.dtype[np.int16]]
ArrayUint16 = np.ndarray[tuple[int], np.dtype[np.uint16]]
ArrayInt32 = np.ndarray[tuple[int], np.dtype[np.int32]]
ArrayUint32 = np.ndarray[tuple[int], np.dtype[np.uint32]]
ArrayInt64 = np.ndarray[tuple[int], np.dtype[np.int64]]
ArrayUint64 = np.ndarray[tuple[int], np.dtype[np.uint64]]

ArrayInt = (
    ArrayInt8
    | ArrayUint8
    | ArrayInt16
    | ArrayUint16
    | ArrayInt32
    | ArrayUint32
    | ArrayInt64
    | ArrayUint64
)

ArrayBool = np.ndarray[tuple[int], np.dtype[np.bool_]]
ArrayDatetime64 = np.ndarray[tuple[int], np.dtype[np.datetime64]]

ArrayStr = np.ndarray[tuple[int], np.dtype[np.str_]]

Array = np.ndarray[tuple[int], Any]  # General array type

TwoDArrayFloat16 = np.ndarray[tuple[int, int], np.dtype[np.float16]]
TwoDArrayFloat32 = np.ndarray[tuple[int, int], np.dtype[np.float32]]
TwoDArrayFloat64 = np.ndarray[tuple[int, int], np.dtype[np.float64]]
TwoDArrayFloat = TwoDArrayFloat16 | TwoDArrayFloat32 | TwoDArrayFloat64

TwoDArrayInt8 = np.ndarray[tuple[int, int], np.dtype[np.int8]]
TwoDArrayInt16 = np.ndarray[tuple[int, int], np.dtype[np.int16]]
TwoDArrayInt32 = np.ndarray[tuple[int, int], np.dtype[np.int32]]
TwoDArrayInt64 = np.ndarray[tuple[int, int], np.dtype[np.int64]]
TwoDArrayUint8 = np.ndarray[tuple[int, int], np.dtype[np.uint8]]
TwoDArrayUint16 = np.ndarray[tuple[int, int], np.dtype[np.uint16]]
TwoDArrayUint32 = np.ndarray[tuple[int, int], np.dtype[np.uint32]]
TwoDArrayUint64 = np.ndarray[tuple[int, int], np.dtype[np.uint64]]
TwoDArrayInt = (
    TwoDArrayInt8
    | TwoDArrayInt16
    | TwoDArrayInt32
    | TwoDArrayInt64
    | TwoDArrayUint8
    | TwoDArrayUint16
    | TwoDArrayUint32
    | TwoDArrayUint64
)

TwoDArrayBool = np.ndarray[tuple[int, int], np.dtype[np.bool_]]

TwoDArray = np.ndarray[tuple[int, int], Any]  # General 2D array type

ThreeDArrayFloat16 = np.ndarray[tuple[int, int, int], np.dtype[np.float16]]
ThreeDArrayFloat32 = np.ndarray[tuple[int, int, int], np.dtype[np.float32]]
ThreeDArrayFloat64 = np.ndarray[tuple[int, int, int], np.dtype[np.float64]]
ThreeDArrayFloat = ThreeDArrayFloat16 | ThreeDArrayFloat32 | ThreeDArrayFloat64

ThreeDArrayInt8 = np.ndarray[tuple[int, int, int], np.dtype[np.int8]]
ThreeDArrayInt16 = np.ndarray[tuple[int, int, int], np.dtype[np.int16]]
ThreeDArrayInt32 = np.ndarray[tuple[int, int, int], np.dtype[np.int32]]
ThreeDArrayInt64 = np.ndarray[tuple[int, int, int], np.dtype[np.int64]]
ThreeDArrayUint8 = np.ndarray[tuple[int, int, int], np.dtype[np.uint8]]
ThreeDArrayUint16 = np.ndarray[tuple[int, int, int], np.dtype[np.uint16]]
ThreeDArrayUint32 = np.ndarray[tuple[int, int, int], np.dtype[np.uint32]]
ThreeDArrayUint64 = np.ndarray[tuple[int, int, int], np.dtype[np.uint64]]
ThreeDArrayInt = (
    ThreeDArrayInt8
    | ThreeDArrayInt16
    | ThreeDArrayInt32
    | ThreeDArrayInt64
    | ThreeDArrayUint8
    | ThreeDArrayUint16
    | ThreeDArrayUint32
    | ThreeDArrayUint64
)

ThreeDArrayBool = np.ndarray[tuple[int, int, int], np.dtype[np.bool_]]

ThreeDArray = np.ndarray[tuple[int, int, int], Any]  # General 3D array type


T_Array = TypeVar("T_Array", bound=Array)
T_TwoDArray = TypeVar("T_TwoDArray", bound=TwoDArray)
T_ThreeDArray = TypeVar("T_ThreeDArray", bound=ThreeDArray)

DType = TypeVar("DType", bound=DTypeLike)
T_OneorTwoDArray = TypeVar("T_OneorTwoDArray", bound=Array | TwoDArray)
T_Scalar = TypeVar("T_Scalar", bound=np.generic)
T_ArrayNumber = TypeVar("T_ArrayNumber", bound=ArrayFloat | ArrayInt)

# Aliases for compress overloads using T_Scalar
ArrayWithScalar = np.ndarray[tuple[int], np.dtype[T_Scalar]]
TwoDArrayWithScalar = np.ndarray[tuple[int, int], np.dtype[T_Scalar]]
ThreeDArrayWithScalar = np.ndarray[tuple[int, int, int], np.dtype[T_Scalar]]
AnyDArrayWithScalar = np.ndarray[Shape, np.dtype[T_Scalar]]



"""Utilities to download OpenBuildingMap tiles for a given bounding box.

This module provides a downloader that downloads and extracts the needed
gpkg tiles from the remote bz2 packages hosted by OpenBuildingMap.
Tiles are quadkey level 6 and downloaded individually.

Notes:
    - OpenBuildingMap distributes quadkey level 6 tiles as individual bz2 (zipped) files.
      Tile filenames follow the pattern " building.002232.gpkg.bz2" where the integers
      indicate the quadkey of the tile. See the OpenBuildingMap documentation for details.
    - Coverage spans the global land areas.

"""

from __future__ import annotations

import bz2
import tempfile
from pathlib import Path
from typing import Any

import geopandas as gpd
import mercantile
import pandas as pd
from pyquadkey2 import quadkey
from shapely import geometry
from tqdm import tqdm

from geb.workflows.io import fetch_and_save, write_geom

from .base import Adapter


class OpenBuildingMap(Adapter):
    """Dataset adapter for OpenBuildingMap data."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the adapter for OpenBuildingMap.

        Args:
            *args: Additional positional arguments passed to the base Adapter class.
            **kwargs: Additional keyword arguments passed to the base Adapter class.
        """
        super().__init__(*args, **kwargs)

    def _quadkeys_for_geom(
        self, geom: geometry.polygon.Polygon, zoom: int = 6
    ) -> list[str]:
        """Gets the quadkeys of tiles that intersect with the polygon geometry.

        Args:
            geom: Polygon geometry for which to get intersecting quadkeys.
            zoom: Zoom level of the quadkeys. Zoomlevel 6 is used for open building map.

        Returns:
            A list of quadkey strings intersecting the polygon.
        """
        quadkeys: list[str] = []

        west, south, east, north = geom.bounds

        # iterate over tiles intersecting the bbox
        for tile in mercantile.tiles(west, south, east, north, zooms=zoom):
            qk = quadkey.from_tile((tile.x, tile.y), level=zoom)

            # Create a polygon for the tile bounds
            tile_bounds = mercantile.bounds(tile)
            tile_polygon = geometry.box(
                tile_bounds.west, tile_bounds.south, tile_bounds.east, tile_bounds.north
            )

            # Only include tile if it intersects with the input geometry
            if tile_polygon.intersects(geom):
                quadkeys.append(qk.key)

        return quadkeys

    def _extract_buildings_in_geom(
        self, gpkg_filename: Path, geom: geometry.polygon.Polygon
    ) -> gpd.GeoDataFrame | None:
        """This function reads the downloaded geopackage containing the buildings. It then extracts only the buildings that lie within the geom.

        Args:
            gpkg_filename: filename of the dowloaded geopackage.
            geom: geom representing the model region.
        Returns:
            A geopandas geodataframe containing all building within the geom.
        """
        # load buildings (should already be masked by region in this step)
        buildings = gpd.read_file(
            gpkg_filename,
            engine="pyogrio",
            mask=geom,
            columns=["id", "occupancy", "floorspace", "height", "geometry"],
            layer="building",
        )
        # mask buildings to region geom
        buildings = buildings[buildings.intersects(geom)]
        if len(buildings) == 0:
            print("No buildings found in region geom")
            return
        else:
            return buildings

    def _download_and_extract_tile(
        self,
        tile_url: str,
        temp_dir: Path,
        tile_filename: str,
    ) -> Path:
        """Download a tile .bz2 and extract only GeoTIFF files that intersect with the bbox.

        Args:
            tile_url: URL of the tile ZIP file.
            temp_dir: Temporary directory to extract to.
            tile_filename: Filename of the tile ZIP.

        Returns:
            Path to the dowloaded geopackage files.

        Raises:
            RuntimeError: If download or extraction fails after all retries.
        """
        # Tile is available, so download with retry
        zip_path: Path = temp_dir / tile_filename
        success: bool = fetch_and_save(
            tile_url,
            zip_path,
            delay_seconds=1,
            verbose=False,
            show_progress=True,
            double_delay=True,
            max_retries=17,  # will be total of ~day
        )
        if not success:
            raise RuntimeError(f"Failed to download {tile_url}")

        # Extract building.gpkg
        gpkg_filename = str(zip_path).replace(".bz2", "")
        gpkg_filename = gpkg_filename.replace("building.", "building_")
        with bz2.open(zip_path, "rb") as f_in, open(gpkg_filename, "wb") as f_out:
            f_out.write(f_in.read())
        return Path(gpkg_filename)

    def fetch(
        self,
        url: str,
        geom: geometry.polygon.Polygon,
        prefix: str,
    ) -> OpenBuildingMap:
        """Download OpenBuildingMap tiles intersecting a bbox.

        Args:
            url: Base URL of the OpenStreetMap server.
            geom: Polygon of the model region.
            prefix: Prefix for the local storage path.

        Returns:
            The OpenBuildingMap instance.

        Raises:
            RuntimeError: If no buildings can be downloaded for the model region.
        """
        if self.path.exists():
            return self

        # get bounds for geom
        tiles: list = self._quadkeys_for_geom(geom=geom)
        with tempfile.TemporaryDirectory() as temp_dir_str:
            temp_dir: Path = Path(temp_dir_str)
            list_of_buildings_in_geom: list[gpd.GeoDataFrame] = []

            for tile in tqdm(tiles, desc="Downloading OpenBuildingMap tiles"):
                tile_filename = f"building.{tile}.gpkg.bz2"
                tile_url: str = f"{url}/{tile_filename}"

                gpkg_filename: Path = self._download_and_extract_tile(
                    tile_url, temp_dir, tile_filename
                )
                # Add all extracted geopackage files (already filtered during extraction)
                buildings = self._extract_buildings_in_geom(gpkg_filename, geom)
                if buildings is not None:
                    list_of_buildings_in_geom.append(buildings)
        # concatenate all buildings
        buildings_in_geom: gpd.GeoDataFrame = pd.concat(
            list_of_buildings_in_geom, ignore_index=True
        )  # ty:ignore[invalid-assignment]

        # raise error if no buildings are found in model region
        if len(list_of_buildings_in_geom) == 0:
            raise RuntimeError("No OpenBuildingMap features were found in model domain")
        # write to file
        self.path.parent.mkdir(parents=True, exist_ok=True)
        write_geom(buildings_in_geom, self.path)

        return self

    def read(self, **kwargs: Any) -> gpd.GeoDataFrame:
        """Read the OpenBuildingMap data as a GeoDataFrame.

        Args:
            **kwargs: Additional keyword arguments to pass to gpd.read_parquet.

        Returns:
            A GeoDataFrame with the GADM data.
        """
        gdf = Adapter.read(self, **kwargs)
        assert isinstance(gdf, gpd.GeoDataFrame)
        return gdf



"""The Lowder adapter for downloading and processing farm size distribution data."""

from __future__ import annotations

from typing import Any

import pandas as pd

from geb.workflows.io import fetch_and_save

from ..workflows.conversions import (
    COUNTRY_NAME_TO_ISO3,
)
from .base import Adapter


class Lowder(Adapter):
    """The Lowder adapter for downloading and processing farm size distribution data."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the Lowder adapter."""
        super().__init__(*args, **kwargs)

    def fetch(self, url: str) -> Lowder:
        """Fetch the Lowder farm size distribution data from the given URL.

        Args:
            url: The URL to download the Lowder farm size distribution data from.

        Returns:
            The instance of the Adapter after fetching the data.
        """
        if not self.is_ready:
            fetch_and_save(url=url, file_path=self.path)
        return self

    def read(self, **kwargs: Any) -> pd.DataFrame:
        """Read and process the Lowder farm size distribution data.

        Args:
            **kwargs: Additional keyword arguments to pass to pandas.read_excel.

        Returns:
            A pandas DataFrame containing the processed farm size distribution data.
        """
        df = super().read(
            sheet_name="WEB table 3",
            skiprows=4,
            skipfooter=2,
            header=None,
            names=[
                "Country",
                "Census Year",
                "Holdings/ agricultural area",
                "Total",
                "< 1 Ha",
                "1 - 2 Ha",
                "2 - 5 Ha",
                "5 - 10 Ha",
                "10 - 20 Ha",
                "20 - 50 Ha",
                "50 - 100 Ha",
                "100 - 200 Ha",
                "200 - 500 Ha",
                "500 - 1000 Ha",
                "> 1000 Ha",
                "empty",
                "income class",
            ],
            **kwargs,
        )
        assert isinstance(df, pd.DataFrame)
        df: pd.DataFrame = df.dropna(subset=["Total"], axis=0).drop(
            ["empty", "income class"], axis=1
        )
        df["Country"] = df["Country"].ffill()
        # Remove preceding and trailing white space from country names
        df["Country"] = df["Country"].str.strip()
        df["Census Year"] = df["Country"].ffill()

        df["ISO3"] = df["Country"].map(COUNTRY_NAME_TO_ISO3)

        # Clean up agricultural area strings
        df["Holdings/ agricultural area"] = (
            df["Holdings/ agricultural area"].str.strip().str.replace("  ", " ")
        )

        assert not df["ISO3"].isna().any(), (
            f"Found {df['ISO3'].isna().sum()} countries without ISO3 code"
        )
        return df
